import re


def remove_sql_comments(sql: str) -> str:
    """Remove SQL comments from the given SQL string."""

    # Remove single-line comments (--) and multi-line comments (/* ... */)
    sql = re.sub(r"--.*?(\n|$)", "", sql)
    sql = re.sub(r"/\*.*?\*/", "", sql, flags=re.DOTALL)
    return sql



"""Base class for vis protocol module."""

import json
from typing import Any, Dict, Optional

from dbgpt.util.json_utils import serialize


class Vis:
    """Vis protocol base class."""

    def render_prompt(self) -> Optional[str]:
        """Return the prompt for the vis protocol."""
        return None

    def sync_generate_param(self, **kwargs) -> Optional[Dict[str, Any]]:
        """Generate the parameters required by the vis protocol.

        Display corresponding content using vis protocol

        Args:
            **kwargs:

        Returns:
        vis protocol text
        """
        return kwargs["content"]

    async def generate_param(self, **kwargs) -> Optional[Dict[str, Any]]:
        """Generate the parameters required by the vis protocol.

        Display corresponding content using vis protocol
        Args:
            **kwargs:

        Returns:
        vis protocol text
        """
        return self.sync_generate_param(**kwargs)

    def sync_display(self, **kwargs) -> Optional[str]:
        """Display the content using the vis protocol."""
        content = json.dumps(
            self.sync_generate_param(**kwargs), default=serialize, ensure_ascii=False
        )
        return f"```{self.vis_tag()}\n{content}\n```"

    async def display(self, **kwargs) -> Optional[str]:
        """Display the content using the vis protocol."""
        return self.sync_display(**kwargs)

    @classmethod
    def vis_tag(cls) -> str:
        """Return current vis protocol module tag name."""
        return ""



"""Application Resources for the agent."""

import dataclasses
from abc import abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Type, cast

from dbgpt._private.pydantic import BaseModel
from dbgpt.agent import AgentMessage, ConversableAgent
from dbgpt.util import ParameterDescription
from dbgpt.util.i18n_utils import _

from .base import Resource, ResourceParameters, ResourceType


class AppInfo(BaseModel):
    code: str
    name: str
    desc: str


class AppResource(Resource[ResourceParameters]):
    """AppResource resource class."""

    def __init__(self, name: str, **kwargs):
        """Initialize AppResource resource."""
        self._resource_name = name

    @property
    @abstractmethod
    def app_desc(self):
        """Return the app description."""

    @property
    @abstractmethod
    def app_name(self):
        """Return the app name."""

    @abstractmethod
    async def _start_app(
        self,
        user_input: str,
        sender: ConversableAgent,
        conv_uid: Optional[str] = None,
    ) -> AgentMessage:
        """start the app"""

    @classmethod
    def type(cls) -> ResourceType:
        """Return the resource type."""
        return ResourceType.App

    @property
    def name(self) -> str:
        """Return the resource name."""
        return self._resource_name

    @classmethod
    def _get_app_list(cls) -> List[AppInfo]:
        """Get the current app list"""

    @classmethod
    def resource_parameters_class(cls, **kwargs) -> Type[ResourceParameters]:
        @dataclasses.dataclass
        class _DynAppResourceParameters(ResourceParameters):
            """Application resource class."""

            apps = cls._get_app_list()
            valid_values = [
                {
                    "label": f"{app.name}({app.code})",
                    "key": app.code,
                    "description": app.desc,
                }
                for app in apps
            ]

            app_code: str = dataclasses.field(
                metadata={
                    "help": _("App code"),
                    "valid_values": valid_values,
                },
            )

            @classmethod
            def to_configurations(
                cls,
                parameters: Type["ResourceParameters"],
                version: Optional[str] = None,
                **kwargs,
            ) -> Any:
                """Convert the parameters to configurations."""
                conf: List[ParameterDescription] = cast(
                    List[ParameterDescription], super().to_configurations(parameters)
                )
                version = version or cls._resource_version()
                if version != "v1":
                    return conf
                # Compatible with old version
                for param in conf:
                    if param.param_name == "app_code":
                        return param.valid_values or []
                return []

            @classmethod
            def from_dict(
                cls, data: dict, ignore_extra_fields: bool = True
            ) -> ResourceParameters:
                """Create a new instance from a dictionary."""
                copied_data = data.copy()
                if "app_code" not in copied_data and "value" in copied_data:
                    copied_data["app_code"] = copied_data.pop("value")
                return super().from_dict(
                    copied_data, ignore_extra_fields=ignore_extra_fields
                )

        return _DynAppResourceParameters

    async def get_prompt(
        self,
        *,
        lang: str = "en",
        prompt_type: str = "default",
        question: Optional[str] = None,
        resource_name: Optional[str] = None,
        **kwargs,
    ) -> Tuple[str, Optional[Dict]]:
        """Get the prompt."""
        prompt_template_zh = (
            "{name}：调用此资源与应用 {app_name} 进行交互。"
            "应用 {app_name} 有什么用？{description}"
        )
        prompt_template_en = (
            "{name}：Call this resource to interact with the application {app_name} ."
            "What is the application {app_name} useful for? {description} "
        )
        template = prompt_template_en if lang == "en" else prompt_template_zh

        return (
            template.format(
                name=self.name, app_name=self.app_name, description=self.app_desc
            ),
            None,
        )

    @property
    def is_async(self) -> bool:
        """Return whether the tool is asynchronous."""
        return True

    def execute(self, *args, resource_name: Optional[str] = None, **kwargs) -> Any:
        """Execute the resource."""
        if self.is_async:
            raise RuntimeError("Sync execution is not supported")

    async def async_execute(
        self,
        *args,
        resource_name: Optional[str] = None,
        **kwargs,
    ) -> Any:
        """Execute the tool asynchronously.

        Args:
            *args: The positional arguments.
            resource_name (str, optional): The tool name to be executed(not used for
            specific tool).
            **kwargs: The keyword arguments.
        """
        user_input: Optional[str] = kwargs.get("user_input")
        parent_agent: Optional[ConversableAgent] = kwargs.get("parent_agent")

        if user_input is None:
            raise RuntimeError("AppResource async execution user_input is None")
        if parent_agent is None:
            raise RuntimeError("AppResource async execution parent_agent is None")

        reply_message = await self._start_app(user_input, parent_agent)
        return reply_message.content



"""Load plugins from a directory or a zip file.

This module provides utility functions to load auto_gpt plugins from a directory or a
zip file.
"""

import datetime
import glob
import json
import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from auto_gpt_plugin_template import AutoGPTPluginTemplate

logger = logging.getLogger(__name__)


def inspect_zip_for_modules(zip_path: str, debug: bool = False) -> list[str]:
    """Load the AutoGPTPluginTemplate from a zip file.

    Loader zip plugin file. Native support Auto_gpt_plugin

    Args:
    zip_path (str): Path to the zipfile.
    debug (bool, optional): Enable debug logging. Defaults to False.

    Returns:
    list[str]: The list of module names found or empty list if none were found.
    """
    import zipfile

    result = []
    with zipfile.ZipFile(zip_path, "r") as zfile:
        for name in zfile.namelist():
            if name.endswith("__init__.py") and not name.startswith("__MACOSX"):
                logger.debug(f"Found module '{name}' in the zipfile at: {name}")
                result.append(name)
    if len(result) == 0:
        logger.debug(f"Module '__init__.py' not found in the zipfile @ {zip_path}.")
    return result


def write_dict_to_json_file(data: dict, file_path: str) -> None:
    """Write a dictionary to a JSON file.

    Args:
        data (dict): Dictionary to write.
        file_path (str): Path to the file.
    """
    with open(file_path, "w") as file:
        json.dump(data, file, indent=4)


def create_directory_if_not_exists(directory_path: str) -> bool:
    """Create a directory if it does not exist.

    Args:
        directory_path (str): Path to the directory.
    Returns:
        bool: True if the directory was created, else False.
    """
    if not os.path.exists(directory_path):
        try:
            os.makedirs(directory_path)
            logger.debug(f"Created directory: {directory_path}")
            return True
        except OSError as e:
            logger.warning(f"Error creating directory {directory_path}: {e}")
            return False
    else:
        logger.info(f"Directory {directory_path} already exists")
        return True


def scan_plugin_file(file_path, debug: bool = False) -> List["AutoGPTPluginTemplate"]:
    """Scan a plugin file and load the plugins."""
    from zipimport import zipimporter

    logger.info(f"__scan_plugin_file:{file_path},{debug}")
    loaded_plugins = []
    if moduleList := inspect_zip_for_modules(str(file_path), debug):
        for module in moduleList:
            plugin = Path(file_path)
            module = Path(module)  # type: ignore
            logger.debug(f"Plugin: {plugin} Module: {module}")
            zipped_package = zipimporter(str(plugin))
            zipped_module = zipped_package.load_module(
                str(module.parent)  # type: ignore
            )
            for key in dir(zipped_module):
                if key.startswith("__"):
                    continue
                a_module = getattr(zipped_module, key)
                a_keys = dir(a_module)
                if (
                    "_abc_impl" in a_keys
                    and a_module.__name__ != "AutoGPTPluginTemplate"
                    # and denylist_allowlist_check(a_module.__name__, cfg)
                ):
                    loaded_plugins.append(a_module())
    return loaded_plugins


def scan_plugins(
    plugins_file_path: str, file_name: str = "", debug: bool = False
) -> List["AutoGPTPluginTemplate"]:
    """Scan the plugins directory for plugins and loads them.

    Args:
        cfg (Config): Config instance including plugins config
        debug (bool, optional): Enable debug logging. Defaults to False.

    Returns:
        List[Tuple[str, Path]]: List of plugins.
    """
    loaded_plugins = []
    # Generic plugins
    plugins_path = Path(plugins_file_path)
    if file_name:
        plugin_path = Path(plugins_path, file_name)
        loaded_plugins = scan_plugin_file(plugin_path)
    else:
        for plugin_path in plugins_path.glob("*.zip"):
            loaded_plugins.extend(scan_plugin_file(plugin_path))

    if loaded_plugins:
        logger.info(f"\nPlugins found: {len(loaded_plugins)}\n--------------------")
    for plugin in loaded_plugins:
        logger.info(f"{plugin._name}: {plugin._version} - {plugin._description}")
    return loaded_plugins


def update_from_git(
    download_path: str,
    github_repo: str = "",
    branch_name: str = "main",
    authorization: Optional[str] = None,
):
    """Update plugins from a git repository."""
    import requests

    os.makedirs(download_path, exist_ok=True)
    if github_repo:
        if github_repo.index("github.com") <= 0:
            raise ValueError("Not a correct Github repository address！" + github_repo)
        github_repo = github_repo.replace(".git", "")
        url = github_repo + "/archive/refs/heads/" + branch_name + ".zip"
        plugin_repo_name = github_repo.strip("/").split("/")[-1]
    else:
        url = (
            "https://github.com/eosphoros-ai/DB-GPT-Plugins/archive/refs/heads/main.zip"
        )
        plugin_repo_name = "DB-GPT-Plugins"
    try:
        session = requests.Session()
        headers = {}
        if authorization and len(authorization) > 0:
            headers = {"Authorization": authorization}
        response = session.get(
            url,
            headers=headers,
        )

        if response.status_code == 200:
            plugins_path_path = Path(download_path)
            files = glob.glob(os.path.join(plugins_path_path, f"{plugin_repo_name}*"))
            for file in files:
                os.remove(file)
            now = datetime.datetime.now()
            time_str = now.strftime("%Y%m%d%H%M%S")
            file_name = (
                f"{plugins_path_path}/{plugin_repo_name}-{branch_name}-{time_str}.zip"
            )
            print(file_name)
            with open(file_name, "wb") as f:
                f.write(response.content)
            return plugin_repo_name
        else:
            logger.error(
                f"Update plugins failed，response code：{response.status_code}"
            )
            raise ValueError(f"Download plugin failed: {response.status_code}")
    except Exception as e:
        logger.error("update plugins from git exception!" + str(e))
        raise ValueError("download plugin exception!", e)


def __fetch_from_git(local_path, git_url):
    import git

    logger.info("fetch plugins from git to local path:{}", local_path)
    os.makedirs(local_path, exist_ok=True)
    repo = git.Repo(local_path)
    if repo.is_repo():
        repo.remotes.origin.pull()
    else:
        git.Repo.clone_from(git_url, local_path)

    # if repo.head.is_valid():
    # clone succ， fetch plugins info



"""The utility functions for chatting with the DAG task."""

import json
import traceback
from typing import Any, AsyncIterator, Dict, Optional

from ...interface.llm import ModelInferenceMetrics, ModelOutput
from ...schema.api import (
    ChatCompletionResponse,
    ChatCompletionResponseChoice,
    ChatCompletionResponseStreamChoice,
    ChatMessage,
    UsageInfo,
)
from ..operators.base import BaseOperator
from ..trigger.http_trigger import CommonLLMHttpResponseBody


def is_chat_flow_type(output_obj: Any, is_class: bool = False) -> bool:
    """Check whether the output object is a chat flow type."""
    if is_class:
        return output_obj in (str, CommonLLMHttpResponseBody, ModelOutput)
    else:
        chat_types = (str, CommonLLMHttpResponseBody)
        return isinstance(output_obj, chat_types)


def is_agent_flow_type(output_obj: Any, is_class: bool = False) -> bool:
    """Check whether the output object is a agent flow type."""
    if is_class:
        return output_obj in (str, CommonLLMHttpResponseBody, ModelOutput)
    else:
        chat_types = (str, CommonLLMHttpResponseBody)
        return isinstance(output_obj, chat_types)


async def safe_chat_with_dag_task(
    task: BaseOperator, request: Any, covert_to_str: bool = False
) -> ModelOutput:
    """Chat with the DAG task.

    Args:
        task (BaseOperator): The DAG task to be executed.
        request (Any): The request to be passed to the DAG task.
        covert_to_str (bool, optional): Whether to convert the output to string.

    Returns:
        ModelOutput: The model output, the result is not incremental.
    """
    try:
        finish_reason = None
        usage = None
        metrics = None
        error_code = 0
        text = ""
        thinking_text = ""
        async for output in safe_chat_stream_with_dag_task(
            task, request, False, covert_to_str=covert_to_str
        ):
            finish_reason = output.finish_reason
            usage = output.usage
            metrics = output.metrics
            error_code = output.error_code
            if output.has_text:
                text = output.text
            if output.has_thinking:
                thinking_text = output.thinking_text
        return ModelOutput.build(
            text,
            thinking_text,
            error_code=error_code,
            metrics=metrics,
            usage=usage,
            finish_reason=finish_reason,
        )
    except Exception as e:
        return ModelOutput(error_code=1, text=str(e), incremental=False)


async def safe_chat_stream_with_dag_task(
    task: BaseOperator, request: Any, incremental: bool, covert_to_str: bool = False
) -> AsyncIterator[ModelOutput]:
    """Chat with the DAG task.

    This function is similar to `chat_stream_with_dag_task`, but it will catch the
    exception and return the error message.

    Args:
        task (BaseOperator): The DAG task to be executed.
        request (Any): The request to be passed to the DAG task.
        incremental (bool): Whether the output is incremental.
        covert_to_str (bool, optional): Whether to convert the output to string.

    Yields:
        ModelOutput: The model output.
    """
    try:
        async for output in chat_stream_with_dag_task(
            task, request, incremental, covert_to_str=covert_to_str
        ):
            yield output
    except Exception as e:
        simple_error_msg = str(e)
        if not simple_error_msg:
            simple_error_msg = traceback.format_exc()
        yield ModelOutput(error_code=1, text=simple_error_msg, incremental=incremental)
    finally:
        if task.streaming_operator and task.dag:
            await task.dag._after_dag_end(task.current_event_loop_task_id)


def _is_sse_output(task: BaseOperator) -> bool:
    """Check whether the DAG task is a server-sent event output.

    Args:
        task (BaseOperator): The DAG task.

    Returns:
        bool: Whether the DAG task is a server-sent event output.
    """
    return task.output_format is not None and task.output_format.upper() == "SSE"


async def chat_stream_with_dag_task(
    task: BaseOperator, request: Any, incremental: bool, covert_to_str: bool = False
) -> AsyncIterator[ModelOutput]:
    """Chat with the DAG task.

    Args:
        task (BaseOperator): The DAG task to be executed.
        request (Any): The request to be passed to the DAG task.
        incremental (bool): Whether the output is incremental.
        covert_to_str (bool, optional): Whether to convert the output to string.

    Yields:
        ModelOutput: The model output.
    """
    is_sse = _is_sse_output(task)
    if not task.streaming_operator:
        try:
            result = await task.call(request)
            model_output = parse_single_output(
                result, is_sse, covert_to_str=covert_to_str
            )
            model_output.incremental = incremental
            yield model_output
        except Exception as e:
            simple_error_msg = str(e)
            if not simple_error_msg:
                simple_error_msg = traceback.format_exc()
            yield ModelOutput(
                error_code=1, text=simple_error_msg, incremental=incremental
            )
    else:
        from dbgpt.model.utils.chatgpt_utils import OpenAIStreamingOutputOperator

        if OpenAIStreamingOutputOperator and isinstance(
            task, OpenAIStreamingOutputOperator
        ):
            full_text = ""
            full_thinking_text = ""
            async for output in await task.call_stream(request):
                model_output = parse_openai_output(output)
                # The output of the OpenAI streaming API is incremental
                if model_output.has_thinking:
                    full_thinking_text += model_output.thinking_text
                if model_output.has_text:
                    full_text += model_output.text
                model_output.incremental = incremental
                if not incremental:
                    model_output = ModelOutput.build(
                        full_text,
                        full_thinking_text,
                        error_code=model_output.error_code,
                        usage=model_output.usage,
                        finish_reason=model_output.finish_reason,
                    )
                yield model_output
                if not model_output.success:
                    break
        else:
            full_text = ""
            full_thinking_text = ""
            previous_text = ""
            previous_thinking_text = ""
            async for output in await task.call_stream(request):
                model_output = parse_single_output(
                    output, is_sse, covert_to_str=covert_to_str
                )
                model_output.incremental = incremental
                if task.incremental_output:
                    # Output is incremental, append the text
                    if model_output.has_thinking:
                        full_thinking_text += model_output.thinking_text
                    if model_output.has_text:
                        full_text += model_output.text
                else:
                    # Output is not incremental, last output is the full text
                    if model_output.has_thinking:
                        full_thinking_text = model_output.thinking_text

                    if model_output.has_text:
                        full_text = model_output.text
                if not incremental:
                    # Return the full text
                    model_output = ModelOutput.build(
                        full_text,
                        full_thinking_text,
                        error_code=model_output.error_code,
                        usage=model_output.usage,
                        finish_reason=model_output.finish_reason,
                    )
                else:
                    # Return the incremental text
                    delta_text = full_text[len(previous_text) :]
                    previous_text = (
                        full_text
                        if len(full_text) > len(previous_text)
                        else previous_text
                    )
                    delta_thinking_text = full_thinking_text[
                        len(previous_thinking_text) :
                    ]
                    previous_thinking_text = (
                        full_thinking_text
                        if len(full_thinking_text) > len(previous_thinking_text)
                        else previous_thinking_text
                    )
                    model_output = ModelOutput.build(
                        delta_text,
                        delta_thinking_text,
                        error_code=model_output.error_code,
                        usage=model_output.usage,
                        finish_reason=model_output.finish_reason,
                    )
                yield model_output
                if not model_output.success:
                    break


def parse_single_output(
    output: Any, is_sse: bool, covert_to_str: bool = False
) -> ModelOutput:
    """Parse the single output.

    Args:
        output (Any): The output to parse.
        is_sse (bool): Whether the output is in SSE format.
        covert_to_str (bool, optional): Whether to convert the output to string.
        Defaults to False.

    Returns:
        ModelOutput: The parsed output.
    """
    finish_reason: Optional[str] = None
    usage: Optional[Dict[str, Any]] = None
    metrics: Optional[ModelInferenceMetrics] = None

    if output is None:
        error_code = 1
        text = "The output is None!"
    elif isinstance(output, str):
        if is_sse:
            sse_output = parse_sse_data(output)
            if sse_output is None:
                error_code = 1
                text = "The output is not a SSE format"
            else:
                error_code = 0
                text = sse_output
        else:
            error_code = 0
            text = output
    elif isinstance(output, ModelOutput):
        # error_code = output.error_code
        # text = output.text
        # finish_reason = output.finish_reason
        # usage = output.usage
        # metrics = output.metrics
        return output
    elif isinstance(output, CommonLLMHttpResponseBody):
        error_code = output.error_code
        text = output.text
    elif isinstance(output, dict):
        error_code = 0
        text = json.dumps(output, ensure_ascii=False)
    elif covert_to_str:
        error_code = 0
        text = str(output)
    else:
        error_code = 1
        text = f"The output is not a valid format({type(output)})"
    return ModelOutput(
        error_code=error_code,
        text=text,
        finish_reason=finish_reason,
        usage=usage,
        metrics=metrics,
    )


def parse_openai_output(output: Any) -> ModelOutput:
    """Parse the OpenAI output.

    Args:
        output (Any): The output to parse. It must be a stream format.

    Returns:
        ModelOutput: The parsed output.
    """
    text = ""
    if not isinstance(output, str):
        return ModelOutput(
            error_code=1,
            text="The output is not a stream format",
        )
    if output.strip() == "data: [DONE]" or output.strip() == "data:[DONE]":
        return ModelOutput(error_code=0, text="")
    if not output.startswith("data:"):
        return ModelOutput(
            error_code=1,
            text="The output is not a stream format",
        )

    sse_output = parse_sse_data(output)
    if sse_output is None:
        return ModelOutput(error_code=1, text="The output is not a SSE format")
    json_data = sse_output.strip()
    try:
        dict_data = json.loads(json_data)
    except Exception as e:
        return ModelOutput(
            error_code=1,
            text=f"Invalid JSON data: {json_data}, {e}",
        )
    if "choices" not in dict_data:
        return ModelOutput(
            error_code=1,
            text=dict_data.get("text", "Unknown error"),
        )
    choices = dict_data["choices"]
    finish_reason: Optional[str] = None
    reasoning_content: Optional[str] = None
    if choices:
        choice = choices[0]
        delta_data = ChatCompletionResponseStreamChoice(**choice)
        if delta_data.delta.content:
            text = delta_data.delta.content
        if delta_data.delta.reasoning_content:
            reasoning_content = delta_data.delta.reasoning_content
        finish_reason = delta_data.finish_reason
    return ModelOutput.build(text, reasoning_content, finish_reason=finish_reason)


def parse_sse_data(output: str) -> Optional[str]:
    r"""Parse the SSE data.

    Just keep the data part.

    Examples:
        .. code-block:: python

            from dbgpt.core.awel.util.chat_util import parse_sse_data

            assert parse_sse_data("data: [DONE]") == "[DONE]"
            assert parse_sse_data("data:[DONE]") == "[DONE]"
            assert parse_sse_data("data: Hello") == "Hello"
            assert parse_sse_data("data: Hello\n") == "Hello"
            assert parse_sse_data("data: Hello\r\n") == "Hello"
            assert parse_sse_data("data: Hi, what's up?") == "Hi, what's up?"

    Args:
        output (str): The output.

    Returns:
        Optional[str]: The parsed data.
    """
    if output.startswith("data:"):
        output = output.strip()
        if output.startswith("data: "):
            output = output[6:]
        else:
            output = output[5:]

        return output
    else:
        return None


def _v1_create_completion_response(
    text: str,
    think_text: Optional[str],
    model_name: str,
    stream_id: str,
    usage: Optional[UsageInfo] = None,
):
    choice_data = ChatCompletionResponseChoice(
        index=0,
        message=ChatMessage(
            role="assistant",
            content=text or "",
            reasoning_content=think_text,
        ),
    )
    _content = ChatCompletionResponse(
        id=stream_id,
        choices=[choice_data],
        model=model_name,
        usage=usage or UsageInfo(),
    )
    _content = json.dumps(_content.model_dump(exclude_unset=True), ensure_ascii=False)
    return f"data: {_content}\n\n"



from __future__ import annotations

import logging
import sys
import time
from collections import OrderedDict
from typing import TYPE_CHECKING, List, Optional, Union

if TYPE_CHECKING:
    from dbgpt.core.interface.message import BaseMessage, ModelMessage

logger = logging.getLogger(__name__)


class ProxyTokenizerWrapper:
    def __init__(self) -> None:
        self._support_encoding = True
        self._encoding_model = None

    def count_token(
        self,
        messages: Union[str, BaseMessage, ModelMessage, List[ModelMessage]],
        model_name: Optional[str] = None,
    ) -> int:
        """Count token of given messages

        Args:
            messages (Union[str, BaseMessage, ModelMessage, List[ModelMessage]]):
                messages to count token
            model_name (Optional[str], optional): model name. Defaults to None.

        Returns:
            int: token count, -1 if failed
        """
        if not self._support_encoding:
            logger.warning(
                "model does not support encoding model, can't count token, returning -1"
            )
            return -1
        encoding = self._get_or_create_encoding_model(model_name)
        cnt = 0
        if isinstance(messages, str):
            cnt = len(encoding.encode(messages, disallowed_special=()))
        elif isinstance(messages, BaseMessage):
            cnt = len(encoding.encode(messages.content, disallowed_special=()))
        elif isinstance(messages, ModelMessage):
            cnt = len(encoding.encode(messages.content, disallowed_special=()))
        elif isinstance(messages, list):
            for message in messages:
                cnt += len(encoding.encode(message.content, disallowed_special=()))
        else:
            logger.warning(
                "unsupported type of messages, can't count token, returning -1"
            )
            return -1
        return cnt

    def _get_or_create_encoding_model(self, model_name: Optional[str] = None):
        """Get or create encoding model for given model name
        More detail see: https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
        """
        if self._encoding_model:
            return self._encoding_model
        try:
            import tiktoken

            logger.info(
                "tiktoken installed, using it to count tokens, tiktoken will download "
                "tokenizer from network, also you can download it and put it in the "
                "directory of environment variable TIKTOKEN_CACHE_DIR"
            )
        except ImportError:
            self._support_encoding = False
            logger.warning("tiktoken not installed, cannot count tokens, returning -1")
            return -1
        try:
            if not model_name:
                model_name = "gpt-3.5-turbo"
            self._encoding_model = tiktoken.model.encoding_for_model(model_name)
        except KeyError:
            logger.warning(
                f"{model_name}'s tokenizer not found, using cl100k_base encoding."
            )
            self._encoding_model = tiktoken.get_encoding("cl100k_base")
        return self._encoding_model


class LRUTokenCache:
    """LRU cache implementation based on count and memory size for token counting
    results"""

    def __init__(self, max_size: int = 1000, max_memory_mb: float = 100):
        """
        Initialize LRU cache

        Args:
            max_size: Maximum number of cache entries
            max_memory_mb: Maximum memory usage (MB)
        """
        # Ensure max_size is at least 1
        self.max_size = max(1, max_size)
        self.max_memory_bytes = max_memory_mb * 1024 * 1024  # Convert to bytes
        self.cache = OrderedDict()  # {key: (value, size_in_bytes, last_access_time)}
        self.current_memory = 0  # Current total memory usage (bytes)

    def get(self, key):
        """Get cache item and update its position"""
        if key not in self.cache:
            return None

        # Get cached value and update access time
        value, size, _ = self.cache[key]
        current_time = time.time()

        # Update cache item position (move to most recently used position)
        self.cache.move_to_end(key)
        # Update last access time
        self.cache[key] = (value, size, current_time)

        return value

    def put(self, key, value):
        """
        Add or update a cache item

        Args:
            key: Cache key
            value: Cache value (token count)
        """
        # Estimate memory size for an integer token count
        size_estimate = sys.getsizeof(key) + sys.getsizeof(value)
        current_time = time.time()

        # If key already exists, remove old value's memory usage
        if key in self.cache:
            _, old_size, _ = self.cache[key]
            self.current_memory -= old_size
            # Move to most recently used position
            self.cache.move_to_end(key)

        # If adding new item would exceed memory limit, delete old items until there's
        # enough space
        while (
            self.current_memory + size_estimate > self.max_memory_bytes
            and self.cache
            and len(self.cache) > 0
        ):
            # Make sure the item we're evicting isn't the one we're about to update
            if len(self.cache) == 1 and key in self.cache:
                # If we only have one item and it's the key we're updating,
                # we should just update it instead of trying to evict
                break

            oldest_key, (_, oldest_size, _) = next(iter(self.cache.items()))
            if oldest_key == key:
                # Skip evicting the key we're about to update
                # Move to the next oldest item
                self.cache.move_to_end(oldest_key)
                if len(self.cache) <= 1:
                    break
                oldest_key, (_, oldest_size, _) = next(iter(self.cache.items()))

            self.cache.pop(oldest_key)  # Remove oldest item
            self.current_memory -= oldest_size
            logger.debug(
                f"LRU cache: Removed token count entry '{oldest_key}' due to memory "
                "limit"
            )

        # If reached count limit but not memory limit, remove oldest item
        if len(self.cache) >= self.max_size and key not in self.cache:
            # Check if the cache is not empty before trying to pop
            if self.cache:
                oldest_key, (_, oldest_size, _) = next(iter(self.cache.items()))
                self.cache.pop(oldest_key)
                self.current_memory -= oldest_size
                logger.debug(
                    f"LRU cache: Removed token count entry '{oldest_key}' due to count "
                    "limit"
                )

        # Add new item
        self.cache[key] = (value, size_estimate, current_time)
        self.current_memory += size_estimate

    def clear(self):
        """Clear the cache"""
        self.cache.clear()
        self.current_memory = 0

    def __len__(self):
        """Return the number of items in the cache"""
        return len(self.cache)



import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, Optional, Type, Union

from dbgpt.core import ModelMetadata
from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.model.proxy.llms.proxy_model import ProxyModel, parse_model_request
from dbgpt.util.i18n_utils import _

from ..base import (
    AsyncGenerateStreamFunction,
    GenerateStreamFunction,
    register_proxy_model_adapter,
)
from .chatgpt import OpenAICompatibleDeployModelParameters, OpenAILLMClient

if TYPE_CHECKING:
    from httpx._types import ProxiesTypes
    from openai import AsyncAzureOpenAI, AsyncOpenAI

    ClientType = Union[AsyncAzureOpenAI, AsyncOpenAI]

_YI_DEFAULT_MODEL = "yi-lightning"


@auto_register_resource(
    label=_("Yi Proxy LLM"),
    category=ResourceCategory.LLM_CLIENT,
    tags={"order": TAGS_ORDER_HIGH},
    description=_("Yi proxy LLM configuration."),
    documentation_url="https://platform.lingyiwanwu.com/docs",
    show_in_ui=False,
)
@dataclass
class YiDeployModelParameters(OpenAICompatibleDeployModelParameters):
    """Deploy model parameters for Yi."""

    provider: str = "proxy/yi"

    api_base: Optional[str] = field(
        default="${env:YI_API_BASE:-https://api.lingyiwanwu.com/v1}",
        metadata={
            "help": _("The base url of the Yi API."),
        },
    )

    api_key: Optional[str] = field(
        default="${env:YI_API_KEY}",
        metadata={
            "help": _("The API key of the Yi API."),
            "tags": "privacy",
        },
    )


async def yi_generate_stream(
    model: ProxyModel, tokenizer, params, device, context_len=2048
):
    client: YiLLMClient = model.proxy_llm_client
    request = parse_model_request(params, client.default_model, stream=True)
    async for r in client.generate_stream(request):
        yield r


class YiLLMClient(OpenAILLMClient):
    """Yi LLM Client.

    Yi' API is compatible with OpenAI's API, so we inherit from OpenAILLMClient.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        api_type: Optional[str] = None,
        api_version: Optional[str] = None,
        model: Optional[str] = _YI_DEFAULT_MODEL,
        proxies: Optional["ProxiesTypes"] = None,
        timeout: Optional[int] = 240,
        model_alias: Optional[str] = _YI_DEFAULT_MODEL,
        context_length: Optional[int] = None,
        openai_client: Optional["ClientType"] = None,
        openai_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        api_base = (
            api_base or os.getenv("YI_API_BASE") or "https://api.lingyiwanwu.com/v1"
        )
        api_key = api_key or os.getenv("YI_API_KEY")
        model = model or _YI_DEFAULT_MODEL
        if not context_length:
            if "200k" in model:
                context_length = 200 * 1024
            else:
                context_length = 4096

        if not api_key:
            raise ValueError(
                "Yi API key is required, please set 'YI_API_KEY' in environment "
                "variable or pass it to the client."
            )
        super().__init__(
            api_key=api_key,
            api_base=api_base,
            api_type=api_type,
            api_version=api_version,
            model=model,
            proxies=proxies,
            timeout=timeout,
            model_alias=model_alias,
            context_length=context_length,
            openai_client=openai_client,
            openai_kwargs=openai_kwargs,
            **kwargs,
        )

    @property
    def default_model(self) -> str:
        model = self._model
        if not model:
            model = _YI_DEFAULT_MODEL
        return model

    @classmethod
    def param_class(cls) -> Type[YiDeployModelParameters]:
        return YiDeployModelParameters

    @classmethod
    def generate_stream_function(
        cls,
    ) -> Optional[Union[GenerateStreamFunction, AsyncGenerateStreamFunction]]:
        return yi_generate_stream


register_proxy_model_adapter(
    YiLLMClient,
    supported_models=[
        ModelMetadata(
            model=["yi-lightning"],
            context_length=16 * 1024,
            description="Yi Lightning by Lingyiwanwu",
            link="https://platform.lingyiwanwu.com/docs",
            function_calling=True,
        ),
    ],
)



"""Module of chat history."""

from .chat_history_db import (  # noqa: F401
    ChatHistoryDao,
    ChatHistoryEntity,
    ChatHistoryMessageEntity,
)
from .storage_adapter import (  # noqa: F401
    DBMessageStorageItemAdapter,
    DBStorageConversationItemAdapter,
)

__ALL__ = [
    "ChatHistoryEntity",
    "ChatHistoryMessageEntity",
    "ChatHistoryDao",
    "DBStorageConversationItemAdapter",
    "DBMessageStorageItemAdapter",
]



"""Core Module for the Agent.

There are four modules in DB-GPT agent core according the paper
`A survey on large language model based autonomous agents
<https://link.springer.com/article/10.1007/s11704-024-40231-1>`
by `Lei Wang, Chen Ma, Xueyang Feng, et al.`:

1. Profiling Module: The profiling module aims to indicate the profiles of the agent
roles.

2. Memory Module: It stores information perceived from the environment and leverages
the recorded memories to facilitate future actions.

3. Planning Module: When faced with a complex task, humans tend to deconstruct it into
simpler subtasks and solve them individually. The planning module aims to empower the
agents with such human capability, which is expected to make the agent behave more
reasonably, powerfully, and reliably

4. Action Module: The action module is responsible for translating the agent’s
decisions into specific outcomes. This module is located at the most downstream
position and directly interacts with the environment.
"""



import pytest

from .. import (
    DAG,
    InputOperator,
    WorkflowRunner,
)


def _register_dag_to_fastapi_app(dag):
    # TODO
    pass


@pytest.mark.asyncio
async def test_http_operator(runner: WorkflowRunner, stream_input_node: InputOperator):
    with DAG("test_map") as dag:
        pass
    _register_dag_to_fastapi_app(dag)



"""Example: Using SkillsMiddleware with MiddlewareAgent.

This example demonstrates how to use the new middleware system
to load and use skills in DB-GPT agents.
"""

import asyncio

from dbgpt.agent.core.agent import AgentContext
from dbgpt.agent.core.profile.base import ProfileConfig
from dbgpt.agent.middleware.agent import AgentConfig, MiddlewareAgent
from dbgpt.agent.skill.middleware_v2 import SkillsMiddlewareV2


async def example_with_skills():
    """Example of using skills with MiddlewareAgent."""

    profile = ProfileConfig(
        name="assistant",
        role="AI Assistant",
        goal="Help users with their tasks using available skills.",
    )

    config = AgentConfig(
        enable_middleware=True,
        enable_skills=True,
        skill_sources=[
            "/path/to/skills/user",
            "/path/to/skills/project",
        ],
        skill_auto_load=True,
        skill_auto_match=True,
        skill_inject_to_prompt=True,
    )

    agent = MiddlewareAgent(
        profile=profile,
        agent_config=config,
    )

    agent_context = AgentContext(
        conv_id="test_conv_001",
        language="zh-CN",
    )

    await agent.bind(agent_context).build()

    return agent


async def example_custom_middleware():
    """Example of using custom middleware."""

    from dbgpt.agent.middleware.base import AgentMiddleware

    class LoggingMiddleware(AgentMiddleware):
        """Custom middleware for logging."""

        async def before_generate_reply(self, agent, context, **kwargs):
            """Called before generating a reply."""
            print(f"Before generate reply: {context.message.content}")

        async def after_generate_reply(self, agent, context, reply_message, **kwargs):
            """Called after generating a reply."""
            print(f"After generate reply: {reply_message.content}")

    profile = ProfileConfig(
        name="assistant",
        role="AI Assistant",
    )

    agent = MiddlewareAgent(
        profile=profile,
        agent_config=AgentConfig(enable_middleware=True),
    )

    logging_middleware = LoggingMiddleware()
    agent.register_middleware(logging_middleware)

    return agent


async def example_skills_middleware_direct():
    """Example of using SkillsMiddlewareV2 directly."""

    skills_middleware = SkillsMiddlewareV2(
        sources=["/path/to/skills/user"],
        auto_load=True,
        auto_match=True,
        inject_to_system_prompt=True,
    )

    skills = skills_middleware.load_skills()
    print(f"Loaded {len(skills)} skills:")
    for name, skill in skills.items():
        print(f"  - {name}: {skill.metadata.description}")

    matched_skills = skills_middleware.match_skills("research quantum computing")
    print(f"\nMatched {len(matched_skills)} skills")


if __name__ == "__main__":
    asyncio.run(example_with_skills())
    asyncio.run(example_custom_middleware())
    asyncio.run(example_skills_middleware_direct())



import logging
import threading
import time
from concurrent.futures import Executor, ThreadPoolExecutor
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Tuple

from dbgpt.component import SystemApp
from dbgpt.core.interface.storage import (
    QuerySpec,
    ResourceIdentifier,
    StorageInterface,
    StorageItem,
)
from dbgpt.util.executor_utils import blocking_func_to_async

from ...base import ModelInstance
from ..registry import ModelRegistry

logger = logging.getLogger(__name__)


@dataclass
class ModelInstanceIdentifier(ResourceIdentifier):
    identifier_split: str = field(default="___$$$$___", init=False)
    model_name: str
    host: str
    port: int

    def __post_init__(self):
        """Post init method."""
        if self.model_name is None:
            raise ValueError("model_name is required.")
        if self.host is None:
            raise ValueError("host is required.")
        if self.port is None:
            raise ValueError("port is required.")

        if any(
            self.identifier_split in key
            for key in [self.model_name, self.host, str(self.port)]
            if key is not None
        ):
            raise ValueError(
                f"identifier_split {self.identifier_split} is not allowed in "
                f"model_name, host, port."
            )

    @property
    def str_identifier(self) -> str:
        """Return the string identifier of the identifier."""
        return self.identifier_split.join(
            key
            for key in [
                self.model_name,
                self.host,
                str(self.port),
            ]
            if key is not None
        )

    def to_dict(self) -> Dict:
        """Convert the identifier to a dict.

        Returns:
            Dict: The dict of the identifier.
        """
        return {
            "model_name": self.model_name,
            "host": self.host,
            "port": self.port,
        }


@dataclass
class ModelInstanceStorageItem(StorageItem):
    model_name: str
    host: str
    port: int
    weight: Optional[float] = 1.0
    check_healthy: Optional[bool] = True
    healthy: Optional[bool] = False
    enabled: Optional[bool] = True
    prompt_template: Optional[str] = None
    last_heartbeat: Optional[datetime] = None
    _identifier: ModelInstanceIdentifier = field(init=False)

    def __post_init__(self):
        """Post init method."""
        # Convert last_heartbeat to datetime if it's a timestamp
        if isinstance(self.last_heartbeat, (int, float)):
            self.last_heartbeat = datetime.fromtimestamp(self.last_heartbeat)

        self._identifier = ModelInstanceIdentifier(
            model_name=self.model_name,
            host=self.host,
            port=self.port,
        )

    @property
    def identifier(self) -> ModelInstanceIdentifier:
        return self._identifier

    def merge(self, other: "StorageItem") -> None:
        if not isinstance(other, ModelInstanceStorageItem):
            raise ValueError(f"Cannot merge with {type(other)}")
        self.from_object(other)

    def to_dict(self) -> Dict:
        last_heartbeat = self.last_heartbeat.timestamp()
        return {
            "model_name": self.model_name,
            "host": self.host,
            "port": self.port,
            "weight": self.weight,
            "check_healthy": self.check_healthy,
            "healthy": self.healthy,
            "enabled": self.enabled,
            "prompt_template": self.prompt_template,
            "last_heartbeat": last_heartbeat,
        }

    def from_object(self, item: "ModelInstanceStorageItem") -> None:
        """Build the item from another item."""
        self.model_name = item.model_name
        self.host = item.host
        self.port = item.port
        self.weight = item.weight
        self.check_healthy = item.check_healthy
        self.healthy = item.healthy
        self.enabled = item.enabled
        self.prompt_template = item.prompt_template
        self.last_heartbeat = item.last_heartbeat

    @classmethod
    def from_model_instance(cls, instance: ModelInstance) -> "ModelInstanceStorageItem":
        return cls(
            model_name=instance.model_name,
            host=instance.host,
            port=instance.port,
            weight=instance.weight,
            check_healthy=instance.check_healthy,
            healthy=instance.healthy,
            enabled=instance.enabled,
            prompt_template=instance.prompt_template,
            last_heartbeat=instance.last_heartbeat,
        )

    @classmethod
    def to_model_instance(cls, item: "ModelInstanceStorageItem") -> ModelInstance:
        return ModelInstance(
            model_name=item.model_name,
            host=item.host,
            port=item.port,
            weight=item.weight,
            check_healthy=item.check_healthy,
            healthy=item.healthy,
            enabled=item.enabled,
            prompt_template=item.prompt_template,
            last_heartbeat=item.last_heartbeat,
        )


class StorageModelRegistry(ModelRegistry):
    def __init__(
        self,
        storage: StorageInterface,
        system_app: SystemApp | None = None,
        executor: Optional[Executor] = None,
        heartbeat_interval_secs: float | int = 60,
        heartbeat_timeout_secs: int = 120,
    ):
        super().__init__(system_app)
        self._storage = storage
        self._executor = executor or ThreadPoolExecutor(max_workers=2)
        self.heartbeat_interval_secs = heartbeat_interval_secs
        self.heartbeat_timeout_secs = heartbeat_timeout_secs
        self.heartbeat_thread = threading.Thread(target=self._heartbeat_checker)
        self.heartbeat_thread.daemon = True
        self.heartbeat_thread.start()

    @classmethod
    def from_url(
        cls,
        db_url: str,
        db_name: str,
        engine_args: Optional[Dict[str, Any]] = None,
        try_to_create_db: bool = False,
        **kwargs,
    ) -> "StorageModelRegistry":
        from dbgpt.storage.metadata.db_manager import DatabaseManager, initialize_db
        from dbgpt.storage.metadata.db_storage import SQLAlchemyStorage
        from dbgpt.util.serialization.json_serialization import JsonSerializer

        from .db_storage import ModelInstanceEntity, ModelInstanceItemAdapter

        if engine_args is None:
            engine_args = {
                "pool_size": 5,
                "max_overflow": 10,
                "pool_timeout": 30,
                "pool_recycle": 3600,
                "pool_pre_ping": True,
            }

        db: DatabaseManager = initialize_db(
            db_url, db_name, engine_args, try_to_create_db=try_to_create_db
        )
        storage_adapter = ModelInstanceItemAdapter()
        serializer = JsonSerializer()
        storage = SQLAlchemyStorage(
            db,
            ModelInstanceEntity,
            storage_adapter,
            serializer,
        )
        return cls(storage, **kwargs)

    async def _get_instances_by_model(
        self, model_name: str, host: str, port: int, healthy_only: bool = False
    ) -> Tuple[List[ModelInstanceStorageItem], List[ModelInstanceStorageItem]]:
        query_spec = QuerySpec(conditions={"model_name": model_name})
        # Query all instances of the model
        instances = await blocking_func_to_async(
            self._executor, self._storage.query, query_spec, ModelInstanceStorageItem
        )
        if healthy_only:
            instances = [ins for ins in instances if ins.healthy is True]
        exist_ins = [ins for ins in instances if ins.host == host and ins.port == port]
        return instances, exist_ins

    def _heartbeat_checker(self):
        while True:
            all_instances: List[ModelInstanceStorageItem] = self._storage.query(
                QuerySpec(conditions={}), ModelInstanceStorageItem
            )
            for instance in all_instances:
                if (
                    instance.check_healthy
                    and datetime.now() - instance.last_heartbeat
                    > timedelta(seconds=self.heartbeat_timeout_secs)
                ):
                    instance.healthy = False
                    self._storage.update(instance)
            time.sleep(self.heartbeat_interval_secs)

    async def register_instance(self, instance: ModelInstance) -> bool:
        model_name = instance.model_name.strip()
        host = instance.host.strip()
        port = instance.port
        _, exist_ins = await self._get_instances_by_model(
            model_name, host, port, healthy_only=False
        )
        if exist_ins:
            # Exist instances, just update the instance
            # One exist instance at most
            ins: ModelInstanceStorageItem = exist_ins[0]
            # Update instance
            ins.weight = instance.weight
            ins.healthy = True
            ins.prompt_template = instance.prompt_template
            ins.last_heartbeat = datetime.now()
            await blocking_func_to_async(self._executor, self._storage.update, ins)
        else:
            # No exist instance, save the new instance
            new_inst = ModelInstanceStorageItem.from_model_instance(instance)
            new_inst.healthy = True
            new_inst.last_heartbeat = datetime.now()
            await blocking_func_to_async(self._executor, self._storage.save, new_inst)
        return True

    async def deregister_instance(self, instance: ModelInstance) -> bool:
        """Deregister a model instance.

        If the instance exists, set the instance as unhealthy, nothing to do if the
        instance does not exist.

        Args:
            instance (ModelInstance): The instance to deregister.
        """
        model_name = instance.model_name.strip()
        host = instance.host.strip()
        port = instance.port
        _, exist_ins = await self._get_instances_by_model(
            model_name, host, port, healthy_only=False
        )
        if exist_ins:
            ins = exist_ins[0]
            ins.healthy = False
            if instance.remove_from_registry:
                logger.info(
                    f"Remove instance {model_name}@{host}:{port} from registry."
                )
                await blocking_func_to_async(
                    self._executor, self._storage.delete, ins.identifier
                )
            else:
                logger.info(f"Set instance {model_name}@{host}:{port} as unhealthy.")
                await blocking_func_to_async(self._executor, self._storage.update, ins)
        return True

    async def get_all_instances(
        self, model_name: str, healthy_only: bool = False
    ) -> List[ModelInstance]:
        """Get all instances of a model(Async).

        Args:
            model_name (str): The model name.
            healthy_only (bool): Whether only get healthy instances. Defaults to False.
        """
        return await blocking_func_to_async(
            self._executor, self.sync_get_all_instances, model_name, healthy_only
        )

    def sync_get_all_instances(
        self, model_name: str, healthy_only: bool = False
    ) -> List[ModelInstance]:
        """Get all instances of a model.

        Args:
            model_name (str): The model name.
            healthy_only (bool): Whether only get healthy instances. Defaults to False.

        Returns:
            List[ModelInstance]: The list of instances.
        """
        instances = self._storage.query(
            QuerySpec(conditions={"model_name": model_name}), ModelInstanceStorageItem
        )
        if healthy_only:
            instances = [ins for ins in instances if ins.healthy is True]
        return [ModelInstanceStorageItem.to_model_instance(ins) for ins in instances]

    async def get_all_model_instances(
        self, healthy_only: bool = False
    ) -> List[ModelInstance]:
        """Get all model instances.

        Args:
            healthy_only (bool): Whether only get healthy instances. Defaults to False.

        Returns:
            List[ModelInstance]: The list of instances.
        """
        all_instances = await blocking_func_to_async(
            self._executor,
            self._storage.query,
            QuerySpec(conditions={}),
            ModelInstanceStorageItem,
        )
        if healthy_only:
            all_instances = [ins for ins in all_instances if ins.healthy is True]
        return [
            ModelInstanceStorageItem.to_model_instance(ins) for ins in all_instances
        ]

    async def send_heartbeat(self, instance: ModelInstance) -> bool:
        """Receive heartbeat from model instance.

        Update the last heartbeat time of the instance. If the instance does not exist,
        register the instance.

        Args:
            instance (ModelInstance): The instance to send heartbeat.

        Returns:
            bool: True if the heartbeat is received successfully.
        """
        model_name = instance.model_name.strip()
        host = instance.host.strip()
        port = instance.port
        _, exist_ins = await self._get_instances_by_model(
            model_name, host, port, healthy_only=False
        )
        if not exist_ins:
            # register new instance from heartbeat
            await self.register_instance(instance)
            return True
        else:
            ins = exist_ins[0]
            ins.last_heartbeat = datetime.now()
            ins.healthy = True
            await blocking_func_to_async(self._executor, self._storage.update, ins)
            return True



"""A proxy agent for the user."""

from typing import List, Optional

from .. import ActionOutput, Agent, AgentMessage
from .base_agent import ConversableAgent
from .profile import ProfileConfig


class UserProxyAgent(ConversableAgent):
    """A proxy agent for the user.

    That can execute code and provide feedback to the other agents.
    """

    profile: ProfileConfig = ProfileConfig(
        name="User",
        role="Human",
        description=(
            "A human admin. Interact with the planner to discuss the plan. "
            "Plan execution needs to be approved by this admin."
        ),
    )

    is_human: bool = True

    ask_user: bool = False

    def have_ask_user(self):
        """If have ask user info in message."""
        return self.ask_user

    async def receive(
        self,
        message: AgentMessage,
        sender: Agent,
        reviewer: Optional[Agent] = None,
        request_reply: Optional[bool] = None,
        silent: Optional[bool] = False,
        is_recovery: Optional[bool] = False,
        is_retry_chat: bool = False,
        last_speaker_name: Optional[str] = None,
        historical_dialogues: Optional[List[AgentMessage]] = None,
        rely_messages: Optional[List[AgentMessage]] = None,
    ) -> None:
        """Receive a message from another agent."""
        if not silent:
            await self._a_process_received_message(message, sender)

        if message.action_report:
            action_report: ActionOutput = message.action_report
            if action_report.ask_user:
                self.ask_user = True



"""Cache utils.

Adapted from https://github.com/hephex/asyncache/blob/master/asyncache/__init__.py.
It has stopped updating since 2022. So I copied the code here for future reference.
"""

import asyncio
import functools
from contextlib import AbstractContextManager
from typing import Any, Callable, MutableMapping, Optional, Protocol, TypeVar

from cachetools import keys

_KT = TypeVar("_KT")
_T = TypeVar("_T")


class IdentityFunction(Protocol):  # pylint: disable=too-few-public-methods
    """
    Type for a function returning the same type as the one it received.
    """

    def __call__(self, __x: _T) -> _T: ...


class NullContext:
    """A class for noop context managers."""

    def __enter__(self):
        """Return ``self`` upon entering the runtime context."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Raise any exception triggered within the runtime context."""
        return None

    async def __aenter__(self):
        """Return ``self`` upon entering the runtime context."""
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Raise any exception triggered within the runtime context."""
        return None


def cached(
    cache: Optional[MutableMapping[_KT, Any]],
    # ignoring the mypy error to be consistent with the type used
    # in https://github.com/python/typeshed/tree/master/stubs/cachetools
    key: Callable[..., _KT] = keys.hashkey,  # type:ignore
    lock: Optional["AbstractContextManager[Any]"] = None,
) -> IdentityFunction:
    """
    Decorator to wrap a function or a coroutine with a memoizing callable
    that saves results in a cache.

    When ``lock`` is provided for a standard function, it's expected to
    implement ``__enter__`` and ``__exit__`` that will be used to lock
    the cache when gets updated. If it wraps a coroutine, ``lock``
    must implement ``__aenter__`` and ``__aexit__``.
    """
    lock = lock or NullContext()

    def decorator(func):
        if asyncio.iscoroutinefunction(func):

            async def wrapper(*args, **kwargs):
                k = key(*args, **kwargs)
                try:
                    async with lock:
                        return cache[k]

                except KeyError:
                    pass  # key not found

                val = await func(*args, **kwargs)

                try:
                    async with lock:
                        cache[k] = val

                except ValueError:
                    pass  # val too large

                return val

        else:

            def wrapper(*args, **kwargs):
                k = key(*args, **kwargs)
                try:
                    with lock:
                        return cache[k]

                except KeyError:
                    pass  # key not found

                val = func(*args, **kwargs)

                try:
                    with lock:
                        cache[k] = val

                except ValueError:
                    pass  # val too large

                return val

        return functools.wraps(func)(wrapper)

    return decorator



"""Embedding retriever."""

from functools import reduce
from typing import Any, Dict, List, Optional, cast

from dbgpt.core import Chunk
from dbgpt.rag.retriever.base import BaseRetriever, RetrieverStrategy
from dbgpt.rag.retriever.rerank import DefaultRanker, Ranker
from dbgpt.rag.retriever.rewrite import QueryRewrite
from dbgpt.storage.base import IndexStoreBase
from dbgpt.storage.vector_store.filters import MetadataFilters
from dbgpt.util.chat_util import run_async_tasks
from dbgpt.util.tracer import root_tracer


class EmbeddingRetriever(BaseRetriever):
    """Embedding retriever."""

    def __init__(
        self,
        index_store: IndexStoreBase,
        top_k: int = 4,
        query_rewrite: Optional[QueryRewrite] = None,
        rerank: Optional[Ranker] = None,
        retrieve_strategy: Optional[RetrieverStrategy] = RetrieverStrategy.EMBEDDING,
    ):
        """Create EmbeddingRetriever.

        Args:
            index_store(IndexStore): vector store connector
            top_k (int): top k
            query_rewrite (Optional[QueryRewrite]): query rewrite
            rerank (Ranker): rerank

        Examples:
            .. code-block:: python

                from dbgpt.storage.vector_store.connector import VectorStoreConnector
                from dbgpt_ext.storage.vector_store.chroma_store import (
                    ChromaVectorConfig,
                )
                from dbgpt.rag.retriever.embedding import EmbeddingRetriever
                from dbgpt.rag.embedding.embedding_factory import (
                    DefaultEmbeddingFactory,
                )

                embedding_factory = DefaultEmbeddingFactory()
                from dbgpt.rag.retriever.embedding import EmbeddingRetriever
                from dbgpt.storage.vector_store.connector import VectorStoreConnector

                embedding_fn = embedding_factory.create(
                    model_name=EMBEDDING_MODEL_CONFIG[CFG.EMBEDDING_MODEL]
                )
                vector_name = "test"
                config = ChromaVectorConfig(name=vector_name, embedding_fn=embedding_fn)
                vector_store_connector = VectorStoreConnector(
                    vector_store_type="Chroma",
                    vector_store_config=config,
                )
                embedding_retriever = EmbeddingRetriever(
                    top_k=3, vector_store_connector=vector_store_connector
                )
                chunks = embedding_retriever.retrieve("your query text")
                print(
                    f"embedding retriever results:{[chunk.content for chunk in chunks]}"
                )
        """
        self._top_k = top_k
        self._query_rewrite = query_rewrite
        self._index_store = index_store
        self._rerank = rerank or DefaultRanker(self._top_k)
        self._retrieve_strategy = retrieve_strategy

    def load_document(self, chunks: List[Chunk], **kwargs: Dict[str, Any]) -> List[str]:
        """Load document in vector database.

        Args:
            chunks (List[Chunk]): document chunks.
        Return:
            List[str]: chunk ids.
        """
        return self._index_store.load_document(chunks)

    def _retrieve(
        self, query: str, filters: Optional[MetadataFilters] = None
    ) -> List[Chunk]:
        """Retrieve knowledge chunks.

        Args:
            query (str): query text
            filters: metadata filters.
        Return:
            List[Chunk]: list of chunks
        """
        queries = [query]
        candidates = [
            self._index_store.similar_search(query, self._top_k, filters)
            for query in queries
        ]
        res_candidates = cast(List[Chunk], reduce(lambda x, y: x + y, candidates))
        return res_candidates

    def _retrieve_with_score(
        self,
        query: str,
        score_threshold: float,
        filters: Optional[MetadataFilters] = None,
    ) -> List[Chunk]:
        """Retrieve knowledge chunks with score.

        Args:
            query (str): query text
            score_threshold (float): score threshold
            filters: metadata filters.
        Return:
            List[Chunk]: list of chunks with score
        """
        queries = [query]
        candidates_with_score = [
            self._index_store.similar_search_with_scores(
                query, self._top_k, score_threshold, filters
            )
            for query in queries
        ]
        new_candidates_with_score = cast(
            List[Chunk], reduce(lambda x, y: x + y, candidates_with_score)
        )
        new_candidates_with_score = self._rerank.rank(new_candidates_with_score, query)
        return new_candidates_with_score

    async def _aretrieve(
        self, query: str, filters: Optional[MetadataFilters] = None
    ) -> List[Chunk]:
        """Retrieve knowledge chunks.

        Args:
            query (str): query text.
            filters: metadata filters.
        Return:
            List[Chunk]: list of chunks
        """
        queries = [query]
        if self._query_rewrite:
            candidates_tasks = [
                self._similarity_search(
                    query, filters, root_tracer.get_current_span_id()
                )
                for query in queries
            ]
            chunks = await self._run_async_tasks(candidates_tasks)
            context = "\n".join([chunk.content for chunk in chunks])
            new_queries = await self._query_rewrite.rewrite(
                origin_query=query, context=context, nums=1
            )
            queries.extend(new_queries)
        candidates = [
            self._similarity_search(query, filters, root_tracer.get_current_span_id())
            for query in queries
        ]
        new_candidates = await run_async_tasks(tasks=candidates, concurrency_limit=1)
        return new_candidates

    async def _aretrieve_with_score(
        self,
        query: str,
        score_threshold: float,
        filters: Optional[MetadataFilters] = None,
    ) -> List[Chunk]:
        """Retrieve knowledge chunks with score.

        Args:
            query (str): query text
            score_threshold (float): score threshold
            filters: metadata filters.
        Return:
            List[Chunk]: list of chunks with score
        """
        queries = [query]
        if self._query_rewrite:
            with root_tracer.start_span(
                "dbgpt.rag.retriever.embeddings.query_rewrite.similarity_search",
                metadata={"query": query, "score_threshold": score_threshold},
            ):
                candidates_tasks = [
                    self._similarity_search(
                        query, filters, root_tracer.get_current_span_id()
                    )
                    for query in queries
                ]
                chunks = await self._run_async_tasks(candidates_tasks)
                context = "\n".join([chunk.content for chunk in chunks])
            with root_tracer.start_span(
                "dbgpt.rag.retriever.embeddings.query_rewrite.rewrite",
                metadata={"query": query, "context": context, "nums": 1},
            ):
                new_queries = await self._query_rewrite.rewrite(
                    origin_query=query, context=context, nums=1
                )
                queries.extend(new_queries)

        with root_tracer.start_span(
            "dbgpt.rag.retriever.embeddings.similarity_search_with_score",
            metadata={"query": query, "score_threshold": score_threshold},
        ):
            candidates_with_score = [
                self._similarity_search_with_score(
                    query, score_threshold, filters, root_tracer.get_current_span_id()
                )
                for query in queries
            ]
            res_candidates_with_score = await run_async_tasks(
                tasks=candidates_with_score, concurrency_limit=1
            )
            new_candidates_with_score = cast(
                List[Chunk], reduce(lambda x, y: x + y, res_candidates_with_score)
            )

        with root_tracer.start_span(
            "dbgpt.rag.retriever.embeddings.rerank",
            metadata={
                "query": query,
                "score_threshold": score_threshold,
                "rerank_cls": self._rerank.__class__.__name__,
            },
        ):
            new_candidates_with_score = await self._rerank.arank(
                new_candidates_with_score, query
            )
            return new_candidates_with_score

    async def _similarity_search(
        self,
        query,
        filters: Optional[MetadataFilters] = None,
        parent_span_id: Optional[str] = None,
    ) -> List[Chunk]:
        """Similar search."""
        with root_tracer.start_span(
            "dbgpt.rag.retriever.embeddings.similarity_search",
            parent_span_id,
            metadata={
                "query": query,
            },
        ):
            return await self._index_store.asimilar_search(query, self._top_k, filters)

    async def _run_async_tasks(self, tasks) -> List[Chunk]:
        """Run async tasks."""
        candidates = await run_async_tasks(tasks=tasks, concurrency_limit=1)
        candidates = reduce(lambda x, y: x + y, candidates)
        return cast(List[Chunk], candidates)

    async def _similarity_search_with_score(
        self,
        query,
        score_threshold,
        filters: Optional[MetadataFilters] = None,
        parent_span_id: Optional[str] = None,
    ) -> List[Chunk]:
        """Similar search with score."""
        with root_tracer.start_span(
            "dbgpt.rag.retriever.embeddings._do_similarity_search_with_score",
            parent_span_id,
            metadata={
                "query": query,
                "score_threshold": score_threshold,
            },
        ):
            return await self._index_store.asimilar_search_with_scores(
                query, self._top_k, score_threshold, filters
            )

    @classmethod
    def name(cls):
        """Return retriever name."""
        return "embedding_retriever"



"""Tests for _wizard.py — model config step and skip/default flow."""

from __future__ import annotations

from unittest.mock import MagicMock, patch

from dbgpt.cli._profiles import get_profile

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _make_log_ask(*return_values):
    """Return a mock _log whose .ask() yields values in sequence."""
    mock_log = MagicMock()
    mock_log.ask.side_effect = list(return_values)
    return mock_log


# ---------------------------------------------------------------------------
# _ask_model_names
# ---------------------------------------------------------------------------


def test_ask_model_names_returns_defaults_on_enter():
    """User presses Enter → default values from spec returned."""
    from dbgpt.cli._wizard import _ask_model_names

    spec = get_profile("openai")

    mock_log = _make_log_ask("gpt-4o", "text-embedding-3-small")
    with patch("dbgpt.cli._wizard._log", mock_log):
        llm, emb = _ask_model_names(spec)

    assert llm == "gpt-4o"
    assert emb == "text-embedding-3-small"
    mock_log.ask.assert_any_call("LLM model name", default="gpt-4o")
    mock_log.ask.assert_any_call(
        "Embedding model name", default="text-embedding-3-small"
    )


def test_ask_model_names_returns_custom_values():
    """User types custom model names → returned as-is."""
    from dbgpt.cli._wizard import _ask_model_names

    spec = get_profile("openai")

    mock_log = _make_log_ask("my-custom-llm", "my-custom-emb")
    with patch("dbgpt.cli._wizard._log", mock_log):
        llm, emb = _ask_model_names(spec)

    assert llm == "my-custom-llm"
    assert emb == "my-custom-emb"


# ---------------------------------------------------------------------------
# run_setup_wizard — model names integration
# ---------------------------------------------------------------------------


def test_run_setup_wizard_calls_ask_model_names(isolated_dbgpt_home):
    """run_setup_wizard passes llm_model/embedding_model to write_profile_config."""
    from dbgpt.cli._wizard import run_setup_wizard

    with (
        patch("dbgpt.cli._wizard._ask_profile") as mock_ask_profile,
        patch("dbgpt.cli._wizard._ask_api_key", return_value="sk-test"),
        patch(
            "dbgpt.cli._wizard._ask_api_base", return_value="https://api.openai.com/v1"
        ),
        patch(
            "dbgpt.cli._wizard._ask_model_names",
            return_value=("gpt-4o", "text-embedding-3-small"),
        ) as mock_model,
        patch(
            "dbgpt.cli._wizard.write_profile_config",
            return_value=isolated_dbgpt_home / "configs" / "openai.toml",
        ) as mock_write,
        patch("dbgpt.cli._wizard._print_welcome"),
    ):
        spec = get_profile("openai")
        mock_ask_profile.return_value = spec

        run_setup_wizard()

        mock_model.assert_called_once_with(spec)
        mock_write.assert_called_once_with(
            "openai",
            api_key="sk-test",
            activate=True,
            llm_model="gpt-4o",
            embedding_model="text-embedding-3-small",
            api_base="https://api.openai.com/v1",
            embedding_api_key=None,
        )


# ---------------------------------------------------------------------------
# Default (formerly skip) flow
# ---------------------------------------------------------------------------


def test_default_profile_bypasses_all_prompts(isolated_dbgpt_home):
    """default spec → _ask_api_key and _ask_model_names never called."""
    from dbgpt.cli._wizard import run_setup_wizard

    with (
        patch("dbgpt.cli._wizard._ask_profile") as mock_ask_profile,
        patch("dbgpt.cli._wizard._ask_api_key") as mock_api_key,
        patch("dbgpt.cli._wizard._ask_model_names") as mock_model,
        patch(
            "dbgpt.cli._wizard.write_profile_config",
            return_value=isolated_dbgpt_home / "configs" / "default.toml",
        ),
        patch("dbgpt.cli._wizard._print_welcome"),
    ):
        mock_ask_profile.return_value = get_profile("default")

        run_setup_wizard()

        mock_api_key.assert_not_called()
        mock_model.assert_not_called()


# ---------------------------------------------------------------------------
# run_setup_noninteractive — unchanged
# ---------------------------------------------------------------------------


def test_run_setup_noninteractive_unchanged(isolated_dbgpt_home):
    """run_setup_noninteractive does NOT call _ask_model_names."""
    from dbgpt.cli._wizard import run_setup_noninteractive

    with (
        patch("dbgpt.cli._wizard._ask_model_names") as mock_model,
        patch(
            "dbgpt.cli._wizard.write_profile_config",
            return_value=isolated_dbgpt_home / "configs" / "openai.toml",
        ),
    ):
        run_setup_noninteractive("openai", "sk-xxx")

        mock_model.assert_not_called()


# ---------------------------------------------------------------------------
# Custom profile — api_base
# ---------------------------------------------------------------------------


def test_custom_profile_asks_api_base(isolated_dbgpt_home):
    """custom spec → _ask_api_base is called and result passed to write_profile_config."""  # noqa: E501
    from dbgpt.cli._wizard import run_setup_wizard

    with (
        patch("dbgpt.cli._wizard._ask_profile") as mock_ask_profile,
        patch("dbgpt.cli._wizard._ask_api_key", return_value="sk-custom"),
        patch(
            "dbgpt.cli._wizard._ask_api_base", return_value="https://my.proxy.com/v1"
        ) as mock_api_base,
        patch(
            "dbgpt.cli._wizard._ask_model_names",
            return_value=("gpt-4o", "text-embedding-3-small"),
        ),
        patch(
            "dbgpt.cli._wizard.write_profile_config",
            return_value=isolated_dbgpt_home / "configs" / "custom.toml",
        ) as mock_write,
        patch("dbgpt.cli._wizard._print_welcome"),
    ):
        spec = get_profile("custom")
        mock_ask_profile.return_value = spec

        run_setup_wizard()

        mock_api_base.assert_called_once_with(spec)
        mock_write.assert_called_once_with(
            "custom",
            api_key="sk-custom",
            activate=True,
            llm_model="gpt-4o",
            embedding_model="text-embedding-3-small",
            api_base="https://my.proxy.com/v1",
            embedding_api_key=None,
        )


def test_openai_profile_asks_api_base(isolated_dbgpt_home):
    """openai spec → _ask_api_base IS called (users often use proxies)."""
    from dbgpt.cli._wizard import run_setup_wizard

    with (
        patch("dbgpt.cli._wizard._ask_profile") as mock_ask_profile,
        patch("dbgpt.cli._wizard._ask_api_key", return_value="sk-openai"),
        patch(
            "dbgpt.cli._wizard._ask_api_base",
            return_value="https://api.openai.com/v1",
        ) as mock_api_base,
        patch(
            "dbgpt.cli._wizard._ask_model_names",
            return_value=("gpt-4o", "text-embedding-3-small"),
        ),
        patch(
            "dbgpt.cli._wizard.write_profile_config",
            return_value=isolated_dbgpt_home / "configs" / "openai.toml",
        ) as mock_write,
        patch("dbgpt.cli._wizard._print_welcome"),
    ):
        spec = get_profile("openai")
        mock_ask_profile.return_value = spec

        run_setup_wizard()

        mock_api_base.assert_called_once_with(spec)
        mock_write.assert_called_once_with(
            "openai",
            api_key="sk-openai",
            activate=True,
            llm_model="gpt-4o",
            embedding_model="text-embedding-3-small",
            api_base="https://api.openai.com/v1",
            embedding_api_key=None,
        )


def test_kimi_profile_does_not_ask_api_base(isolated_dbgpt_home):
    """kimi spec → _ask_api_base is NOT called."""
    from dbgpt.cli._wizard import run_setup_wizard

    with (
        patch("dbgpt.cli._wizard._ask_profile") as mock_ask_profile,
        patch("dbgpt.cli._wizard._ask_api_key", return_value="sk-test"),
        patch("dbgpt.cli._wizard._ask_api_base") as mock_api_base,
        patch(
            "dbgpt.cli._wizard._ask_model_names",
            return_value=("kimi-k2", "text-embedding-v3"),
        ),
        patch("dbgpt.cli._wizard._ask_embedding_api_key", return_value="ds-key"),
        patch(
            "dbgpt.cli._wizard.write_profile_config",
            return_value=isolated_dbgpt_home / "configs" / "kimi.toml",
        ),
        patch("dbgpt.cli._wizard._print_welcome"),
    ):
        mock_ask_profile.return_value = get_profile("kimi")

        run_setup_wizard()

        mock_api_base.assert_not_called()



"""Tests for MiniMax proxy LLM client."""

import os
from unittest.mock import patch

from dbgpt.model.proxy.llms.minimax import (
    _DEFAULT_MODEL,
    MiniMaxDeployModelParameters,
    MiniMaxLLMClient,
)


class TestMiniMaxDefaults:
    """Test MiniMax default model configuration."""

    def test_default_model_is_m27(self):
        assert _DEFAULT_MODEL == "MiniMax-M2.7"

    @patch.dict(os.environ, {"MINIMAX_API_KEY": "test-key"})
    def test_client_default_model(self):
        client = MiniMaxLLMClient()
        assert client.default_model == "MiniMax-M2.7"

    @patch.dict(os.environ, {"MINIMAX_API_KEY": "test-key"})
    def test_client_custom_model(self):
        client = MiniMaxLLMClient(model="MiniMax-M2.5")
        assert client.default_model == "MiniMax-M2.5"

    def test_deploy_params_provider(self):
        assert MiniMaxDeployModelParameters.provider == "proxy/minimax"


class TestMiniMaxModelList:
    """Test MiniMax model registration."""

    def test_model_list_contains_m27(self):
        # Import triggers registration
        from dbgpt.model.adapter.base import get_model_adapter

        adapter = get_model_adapter("proxy/minimax", "MiniMax-M2.7")
        assert adapter is not None

    def test_model_list_contains_m27_highspeed(self):
        from dbgpt.model.adapter.base import get_model_adapter

        adapter = get_model_adapter("proxy/minimax", "MiniMax-M2.7-highspeed")
        assert adapter is not None

    def test_model_list_contains_m25(self):
        from dbgpt.model.adapter.base import get_model_adapter

        adapter = get_model_adapter("proxy/minimax", "MiniMax-M2.5")
        assert adapter is not None

    def test_model_list_contains_m25_highspeed(self):
        from dbgpt.model.adapter.base import get_model_adapter

        adapter = get_model_adapter("proxy/minimax", "MiniMax-M2.5-highspeed")
        assert adapter is not None


class TestMiniMaxTemperatureClamping:
    """Test temperature clamping behavior."""

    @patch.dict(os.environ, {"MINIMAX_API_KEY": "test-key"})
    def test_valid_temperature_passthrough(self):
        from dbgpt.core import ModelMessage, ModelRequest

        client = MiniMaxLLMClient()
        request = ModelRequest(
            model="MiniMax-M2.7",
            messages=[ModelMessage(role="user", content="hi")],
            temperature=0.5,
        )
        payload = client._build_request(request)
        assert payload["temperature"] == 0.5

    @patch.dict(os.environ, {"MINIMAX_API_KEY": "test-key"})
    def test_high_temperature_clamped(self):
        from dbgpt.core import ModelMessage, ModelRequest

        client = MiniMaxLLMClient()
        request = ModelRequest(
            model="MiniMax-M2.7",
            messages=[ModelMessage(role="user", content="hi")],
            temperature=2.0,
        )
        payload = client._build_request(request)
        assert payload["temperature"] == 1.0



"""Blank Action for the Agent."""

import logging
from typing import Optional

from ...resource.base import AgentResource
from .base import Action, ActionOutput

logger = logging.getLogger(__name__)


class BlankAction(Action):
    """Blank action class."""

    def __init__(self, **kwargs):
        """Blank action init."""
        super().__init__(**kwargs)

    @property
    def ai_out_schema(self) -> Optional[str]:
        """Return the AI output schema."""
        return None

    async def run(
        self,
        ai_message: str,
        resource: Optional[AgentResource] = None,
        rely_action_out: Optional[ActionOutput] = None,
        need_vis_render: bool = True,
        **kwargs,
    ) -> ActionOutput:
        """Perform the action.

        Just return the AI message.
        """
        return ActionOutput(
            is_exe_success=True,
            content=ai_message,
            view=ai_message,
        )



"""User-level configuration management for DB-GPT CLI.

Manages ``~/.dbgpt/configs/<profile>.toml`` — one flat TOML file per
profile — and a small ``~/.dbgpt/config.toml`` that records which profile
is active by default.
"""

from __future__ import annotations

import os
from pathlib import Path
from typing import Optional

# ---------------------------------------------------------------------------
# Paths
# ---------------------------------------------------------------------------

_DBGPT_HOME = Path(os.environ.get("DBGPT_HOME", str(Path.home() / ".dbgpt")))
_CONFIGS_DIR = _DBGPT_HOME / "configs"
_ACTIVE_CONFIG = _DBGPT_HOME / "config.toml"


def dbgpt_home() -> Path:
    """Return ``~/.dbgpt``, creating it if necessary."""
    _DBGPT_HOME.mkdir(parents=True, exist_ok=True)
    return _DBGPT_HOME


def configs_dir() -> Path:
    """Return ``~/.dbgpt/configs/``, creating it if necessary."""
    _CONFIGS_DIR.mkdir(parents=True, exist_ok=True)
    return _CONFIGS_DIR


def profile_config_path(profile_name: str) -> Path:
    """Return the path for a profile TOML, e.g. ``~/.dbgpt/configs/openai.toml``."""
    return configs_dir() / f"{profile_name}.toml"


def active_config_path() -> Path:
    """Return ``~/.dbgpt/config.toml``."""
    return _ACTIVE_CONFIG


# ---------------------------------------------------------------------------
# Active-profile record
# ---------------------------------------------------------------------------


def read_active_profile() -> Optional[str]:
    """Return the name of the active profile from ``~/.dbgpt/config.toml``.

    Returns:
        Optional[str]: Profile name, or *None* if not yet configured.
    """
    path = active_config_path()
    if not path.exists():
        return None
    try:
        import tomlkit  # type: ignore[import]

        data = tomlkit.loads(path.read_text(encoding="utf-8"))
        return data.get("default", {}).get("profile") or None
    except Exception:
        return None


def write_active_profile(profile_name: str) -> None:
    """Persist the active profile name to ``~/.dbgpt/config.toml``.

    Args:
        profile_name (str): The profile to activate.
    """
    import tomlkit  # type: ignore[import]

    path = active_config_path()
    dbgpt_home()  # ensure directory exists

    if path.exists():
        try:
            data = tomlkit.loads(path.read_text(encoding="utf-8"))
        except Exception:
            data = tomlkit.document()
    else:
        data = tomlkit.document()

    if "default" not in data:
        data["default"] = tomlkit.table()
    data["default"]["profile"] = profile_name  # type: ignore[index]
    path.write_text(tomlkit.dumps(data), encoding="utf-8")


# ---------------------------------------------------------------------------
# Profile TOML generation
# ---------------------------------------------------------------------------


def _escape_toml_string(value: str) -> str:
    """Escape backslashes and double quotes for TOML basic strings."""
    return value.replace("\\", "\\\\").replace('"', '\\"')


def _render_profile_toml(
    spec: "ProfileSpec",  # noqa: F821
    api_key: Optional[str],
    llm_model: Optional[str] = None,
    embedding_model: Optional[str] = None,
    api_base: Optional[str] = None,
    embedding_api_key: Optional[str] = None,
) -> str:
    """Render a complete TOML config for the given profile.

    The generated file uses a literal API key when *api_key* is provided;
    otherwise it uses the ``${env:VAR}`` interpolation syntax so the server
    reads the key from the environment at runtime.

    Args:
        spec: A :class:`~dbgpt.cli._profiles.ProfileSpec` instance.
        api_key (Optional[str]): Literal API key value, or *None*.
        llm_model (Optional[str]): Override for the LLM model name. If *None*,
            uses ``spec.llm_model``.
        embedding_model (Optional[str]): Override for the embedding model name.
            If *None*, uses ``spec.embedding_model``.
        api_base (Optional[str]): Override for the LLM API base URL. If *None*,
            uses ``spec.llm_api_base``.
        embedding_api_key (Optional[str]): Literal embedding API key, or
            *None*.  When *None* and ``spec.embedding_env_var`` is set (and
            differs from ``spec.env_var``), the generated TOML will reference
            that separate env var instead of the LLM key.

    Returns:
        str: TOML content as a string.
    """
    from dbgpt.cli._profiles import ProfileSpec  # noqa: F401 (type-only)

    if api_key:
        api_key_value = _escape_toml_string(api_key)
    elif spec.env_var:
        api_key_value = f"${{env:{spec.env_var}:-sk-xxx}}"
    else:
        api_key_value = ""

    emb_env_var = spec.embedding_env_var or spec.env_var
    if embedding_api_key:
        emb_key_value = _escape_toml_string(embedding_api_key)
    elif api_key and not spec.embedding_env_var:
        emb_key_value = _escape_toml_string(api_key)
    elif emb_env_var:
        emb_key_value = f"${{env:{emb_env_var}:-sk-xxx}}"
    else:
        emb_key_value = ""

    if api_base:
        embedding_api_url = f"{api_base.rstrip('/')}/embeddings"
    else:
        embedding_api_url = spec.embedding_api_url

    data_dir = str(dbgpt_home() / "workspace" / "pilot" / "meta_data" / "dbgpt.db")
    vector_dir = str(dbgpt_home() / "workspace" / "pilot" / "data")

    lines = [
        f"# DB-GPT configuration — profile: {spec.name}",
        "# Generated by `dbgpt setup`",
        "",
        "[system]",
        "# Load language from environment variable(It is set by the hook)",
        'language = "${env:DBGPT_LANG:-en}"',
        "api_keys = []",
        'encrypt_key = "your_secret_key"',
        "",
        "# Server Configurations",
        "[service.web]",
        'host = "0.0.0.0"',
        "port = 5670",
        "",
        "[service.web.database]",
        'type = "sqlite"',
        f'path = "{data_dir}"',
        "",
        "[rag.storage]",
        "[rag.storage.vector]",
        'type = "chroma"',
        f'persist_path = "{vector_dir}"',
        "",
        "# Model Configurations",
        "[models]",
        "[[models.llms]]",
    ]

    use_env = getattr(spec, "use_env_interpolation", False) and not api_key

    if use_env:
        lines.append(f'name = "${{env:LLM_MODEL_NAME:-{spec.llm_model}}}"')
        lines.append(f'provider = "${{env:LLM_MODEL_PROVIDER:-{spec.llm_provider}}}"')
        effective_api_base = api_base or spec.llm_api_base
        if effective_api_base:
            lines.append(f'api_base = "${{env:OPENAI_API_BASE:-{effective_api_base}}}"')
        lines.append(f'api_key = "{api_key_value}"')
        lines.append("")
        lines.append("[[models.embeddings]]")
        lines.append(f'name = "${{env:EMBEDDING_MODEL_NAME:-{spec.embedding_model}}}"')
        lines.append(
            f'provider = "${{env:EMBEDDING_MODEL_PROVIDER:-{spec.embedding_provider}}}"'
        )
        lines.append(
            f'api_url = "${{env:EMBEDDING_MODEL_API_URL:-{embedding_api_url}}}"'
        )
        lines.append(f'api_key = "{emb_key_value}"')
    else:
        lines.append(f'name = "{llm_model or spec.llm_model}"')
        lines.append(f'provider = "{spec.llm_provider}"')

        effective_api_base = api_base or spec.llm_api_base
        if effective_api_base:
            lines.append(f'api_base = "{effective_api_base}"')

        lines.append(f'api_key = "{api_key_value}"')
        lines.append("")
        lines.append("[[models.embeddings]]")
        lines.append(f'name = "{embedding_model or spec.embedding_model}"')
        lines.append(f'provider = "{spec.embedding_provider}"')
        lines.append(f'api_url = "{embedding_api_url}"')
        lines.append(f'api_key = "{emb_key_value}"')

    # Append any provider-specific extras
    for extra in spec.extra_toml_lines:
        lines.append(extra)

    return "\n".join(lines) + "\n"


# ---------------------------------------------------------------------------
# Public write API
# ---------------------------------------------------------------------------


def write_profile_config(
    profile_name: str,
    api_key: Optional[str] = None,
    activate: bool = True,
    llm_model: Optional[str] = None,
    embedding_model: Optional[str] = None,
    api_base: Optional[str] = None,
    embedding_api_key: Optional[str] = None,
) -> Path:
    """Write (or overwrite) the TOML config for *profile_name*.

    Args:
        profile_name (str): One of the supported profile names.
        api_key (Optional[str]): Literal API key.  If *None*, env-var
            interpolation is used instead.
        activate (bool): Also update ``~/.dbgpt/config.toml`` to make this
            profile the active default.
        llm_model (Optional[str]): Override LLM model name. Uses spec default
            if *None*.
        embedding_model (Optional[str]): Override embedding model name. Uses
            spec default if *None*.
        api_base (Optional[str]): Override LLM API base URL. Uses spec default
            if *None*.
        embedding_api_key (Optional[str]): Literal embedding API key. When
            *None* and the profile has a separate ``embedding_env_var``, the
            generated TOML will reference that env var.

    Returns:
        Path: Path to the written config file.
    """
    from dbgpt.cli._profiles import get_profile

    spec = get_profile(profile_name)
    content = _render_profile_toml(
        spec,
        api_key,
        llm_model=llm_model,
        embedding_model=embedding_model,
        api_base=api_base,
        embedding_api_key=embedding_api_key,
    )
    path = profile_config_path(profile_name)
    path.write_text(content, encoding="utf-8")

    if activate:
        write_active_profile(profile_name)

    return path


def resolve_config_path(
    profile: Optional[str] = None,
    config: Optional[str] = None,
) -> Optional[str]:
    """Resolve which config file to use, in priority order.

    Priority:
    1. Explicit ``--config`` flag → use as-is.
    2. Explicit ``--profile`` flag → look up ``~/.dbgpt/configs/<profile>.toml``.
    3. Active profile from ``~/.dbgpt/config.toml``.
    4. Return *None* (caller should run the setup wizard).

    Args:
        profile (Optional[str]): Value of ``--profile`` CLI flag.
        config (Optional[str]): Value of ``--config`` CLI flag.

    Returns:
        Optional[str]: Absolute path to the config file, or *None*.
    """
    if config:
        return config

    if profile:
        path = profile_config_path(profile)
        if path.exists():
            return str(path)
        return None  # profile specified but not yet configured

    # Fall back to whatever is active
    active = read_active_profile()
    if active:
        path = profile_config_path(active)
        if path.exists():
            return str(path)

    return None



from typing import Generic, List, TypeVar

from dbgpt._private.pydantic import BaseModel, ConfigDict, Field

T = TypeVar("T")


class PaginationResult(BaseModel, Generic[T]):
    """Pagination result"""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    items: List[T] = Field(..., description="The items in the current page")
    total_count: int = Field(..., description="Total number of items")
    total_pages: int = Field(..., description="total number of pages")
    page: int = Field(..., description="Current page number")
    page_size: int = Field(..., description="Number of items per page")

    @classmethod
    def build_from_all(
        cls, all_items: List[T], page: int, page_size: int
    ) -> "PaginationResult[T]":
        """Build a pagination result from all items"""
        if page < 1:
            page = 1
        if page_size < 1:
            page_size = 1
        total_count = len(all_items)
        total_pages = (
            (total_count + page_size - 1) // page_size if total_count > 0 else 0
        )
        page = max(1, min(page, total_pages)) if total_pages > 0 else 0
        start_index = (page - 1) * page_size if page > 0 else 0
        end_index = min(start_index + page_size, total_count)
        items = all_items[start_index:end_index]

        return cls(
            items=items,
            total_count=total_count,
            total_pages=total_pages,
            page=page,
            page_size=page_size,
        )



"""Plan module for the agent."""

from .awel.agent_operator import (  # noqa: F401
    AgentDummyTrigger,
    AWELAgentOperator,
    WrappedAgentOperator,
)
from .awel.agent_operator_resource import (  # noqa: F401
    AWELAgent,
    AWELAgentConfig,
    AWELAgentResource,
)
from .awel.team_awel_layout import (  # noqa: F401
    AWELTeamContext,
    DefaultAWELLayoutManager,
    WrappedAWELLayoutManager,
)
from .plan_action import PlanAction, PlanInput  # noqa: F401
from .planner_agent import PlannerAgent  # noqa: F401
from .team_auto_plan import AutoPlanChatManager  # noqa: F401

__all__ = [
    "PlanAction",
    "PlanInput",
    "PlannerAgent",
    "AutoPlanChatManager",
    "AWELAgent",
    "AWELAgentConfig",
    "AWELAgentResource",
    "AWELTeamContext",
    "DefaultAWELLayoutManager",
    "WrappedAWELLayoutManager",
    "AgentDummyTrigger",
    "AWELAgentOperator",
    "WrappedAgentOperator",
]



"""Skills middleware using the new middleware system.

This module implements SkillsMiddleware as an AgentMiddleware, following
deepagents' pattern with lifecycle hooks.
"""

import logging
from typing import Any, Dict, List, Optional

from ..middleware.base import AgentMiddleware
from .middleware import LoadedSkill
from .middleware import SkillsMiddleware as BaseSkillsMiddleware

logger = logging.getLogger(__name__)


class SkillsMiddlewareV2(AgentMiddleware):
    """Skills middleware using AgentMiddleware pattern.

    Integrates with the middleware system to inject skills into the agent
    lifecycle, similar to deepagents' SkillsMiddleware.

    Example:
        ```python
        from dbgpt.agent.middleware.base import MiddlewareManager
        from dbgpt.agent.skill.middleware_v2 import SkillsMiddlewareV2

        middleware = SkillsMiddlewareV2(
            sources=["/path/to/skills/user", "/path/to/skills/project"]
        )
        manager = MiddlewareManager()
        manager.register(middleware)
        ```
    """

    def __init__(
        self,
        sources: List[str],
        auto_load: bool = True,
        auto_match: bool = True,
        inject_to_system_prompt: bool = True,
    ):
        """Initialize the skills middleware.

        Args:
            sources: List of skill source paths.
            auto_load: Automatically load skills on initialization.
            auto_match: Automatically match skills based on user input.
            inject_to_system_prompt: Inject skills into system prompt.
        """
        super().__init__()
        self.sources = sources
        self.auto_load = auto_load
        self.auto_match = auto_match
        self.inject_to_system_prompt = inject_to_system_prompt

        self._base_middleware = BaseSkillsMiddleware(sources=sources)
        self._loaded = False

        if auto_load:
            self.load_skills()

    def load_skills(self) -> Dict[str, LoadedSkill]:
        """Load skills from all configured sources.

        Returns:
            Dictionary of loaded skills keyed by name.
        """
        self._loaded = True
        return self._base_middleware.load_skills()

    def get_skill(self, name: str) -> Optional[LoadedSkill]:
        """Get a skill by name.

        Args:
            name: Skill name.

        Returns:
            LoadedSkill or None.
        """
        if not self._loaded:
            self.load_skills()
        return self._base_middleware.get_skill(name)

    def list_skills(self) -> List:
        """List all loaded skills.

        Returns:
            List of skill metadata.
        """
        if not self._loaded:
            self.load_skills()
        return self._base_middleware.list_skills()

    def match_skills(self, user_input: str) -> List[LoadedSkill]:
        """Find skills that match user input.

        Args:
            user_input: User input string.

        Returns:
            List of matching skills.
        """
        if not self._loaded:
            self.load_skills()
        return self._base_middleware.match_skills(user_input)

    async def after_init(self, agent, **kwargs) -> Optional[Dict[str, Any]]:
        """Called after agent initialization.

        Loads skills and stores them in middleware state.

        Args:
            agent: The agent instance.
            **kwargs: Additional arguments.

        Returns:
            State with skills metadata.
        """
        logger.info(f"SkillsMiddlewareV2: after_init for {agent.__class__.__name__}")

        if self.auto_load:
            skills = self.load_skills()
            logger.info(f"Loaded {len(skills)} skills from {self.sources}")
            return {
                "skills_loaded": True,
                "skills_count": len(skills),
                "skills_sources": self.sources,
            }
        return {}

    async def before_generate_reply(
        self,
        agent,
        context,
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """Called before generating a reply.

        Matches skills based on user input if auto_match is enabled.

        Args:
            agent: The agent instance.
            context: The generate context.
            **kwargs: Additional arguments.

        Returns:
            State with matched skills.
        """
        if not self.auto_match or not context.message:
            return {}

        user_input = context.message.content or ""
        if user_input:
            matched_skills = self.match_skills(user_input)
            if matched_skills:
                logger.info(
                    f"Matched {len(matched_skills)} skills for input:"
                    f" {user_input[:50]}..."
                )
                return {
                    "matched_skills": [s.metadata.name for s in matched_skills],
                    "matched_skills_count": len(matched_skills),
                }
        return {}

    async def modify_system_prompt(
        self,
        agent,
        original_prompt: str,
        context: Optional[Dict[str, Any]] = None,
    ) -> str:
        """Modify the system prompt with skills information.

        Args:
            agent: The agent instance.
            original_prompt: The original system prompt.
            context: Additional context information.

        Returns:
            Modified system prompt with skills section.
        """
        if not self.inject_to_system_prompt:
            return original_prompt

        if not self._loaded:
            self.load_skills()

        skills_section = self._base_middleware.create_skills_prompt_section()

        if original_prompt:
            modified = f"{skills_section}\n\n{original_prompt}"
        else:
            modified = skills_section

        return modified

    def get_skills_summary(self) -> str:
        """Get a summary of available skills.

        Returns:
            Formatted summary string.
        """
        all_skills = self.list_skills()
        summary = f"Total Skills: {len(all_skills)}\n"
        summary += f"Sources: {', '.join(self.sources)}\n\n"
        summary += "**Available Skills:**\n"
        for skill_metadata in all_skills:
            summary += f"  - {skill_metadata.name}: {skill_metadata.description}\n"

        return summary


__all__ = [
    "SkillsMiddlewareV2",
]



import logging
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

from dbgpt.core import (
    EmbeddingModelMetadata,
    Embeddings,
    ModelMetadata,
    RerankEmbeddings,
)
from dbgpt.core.interface.media import MediaProcessor
from dbgpt.core.interface.message import ModelMessage, ModelMessageRoleType
from dbgpt.core.interface.parameter import (
    BaseDeployModelParameters,
    LLMDeployModelParameters,
)
from dbgpt.model.adapter.template import (
    ConversationAdapter,
    ConversationAdapterFactory,
    get_conv_template,
)
from dbgpt.model.base import ModelType, SupportedModel
from dbgpt.model.parameter import WorkerType

logger = logging.getLogger(__name__)


class EmbeddingModelAdapter(ABC):
    """New Adapter for DB-GPT Embedding models"""

    model_name: Optional[str] = None
    model_path: Optional[str] = None
    _supported_models: List[EmbeddingModelMetadata] = []

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} model_name={self.model_name} "
            f"model_path={self.model_path}>"
        )

    def new_adapter(self, **kwargs) -> "EmbeddingModelAdapter":
        """Create a new adapter instance

        Args:
            **kwargs: The parameters of the new adapter instance

        Returns:
            EmbeddingModelAdapter: The new adapter instance
        """
        new_obj = self.__class__()
        new_obj.model_name = self.model_name
        new_obj.model_path = self.model_path
        new_obj._supported_models = self._supported_models
        return new_obj

    def model_type(self) -> str:
        return ModelType.HF

    def model_param_class(
        self, model_type: Optional[str] = None
    ) -> Type[BaseDeployModelParameters]:
        """Get the startup parameters instance of the model

        Args:
            model_type (str, optional): The type of model. Defaults to None.

        Returns:
            Type[BaseModelParameters]: The startup parameters instance of the model
        """
        raise NotImplementedError

    def supported_models(self) -> List[EmbeddingModelMetadata]:
        """Return the supported models."""
        return self._supported_models

    def match(
        self,
        provider: str,
        is_rerank: bool,
        model_name: Optional[str] = None,
        model_path: Optional[str] = None,
    ) -> bool:
        """Whether the model adapter can load the given model

        Args:
            provider (str): The provider of the model, old version is model_type.
            model_name (Optional[str], optional): The name of model. Defaults to None.
            model_path (Optional[str], optional): The path of model. Defaults to None.
        """
        model_cls = self.model_param_class()
        if model_cls.provider != provider:
            return False
        if model_name is None and model_path is None:
            return False
        model_name = model_name.lower() if model_name else None
        model_path = model_path.lower() if model_path else None
        return self.do_match(is_rerank, model_name) or self.do_match(
            is_rerank, model_path
        )

    @abstractmethod
    def do_match(self, is_rerank: bool, lower_model_name_or_path: Optional[str] = None):
        """Whether the model adapter can load the given model."""
        raise NotImplementedError()

    def load(
        self, model_path: str, embedding_kwargs: dict
    ) -> Union[Embeddings, RerankEmbeddings]:
        """Load the embeddings model or rerank embeddings model."""
        raise NotImplementedError

    def load_from_params(self, params):
        """Load the model according to the given parameters"""
        raise NotImplementedError

    def support_async(self) -> bool:
        """Whether the loaded model supports asynchronous calls"""
        return False


class LLMModelAdapter(ABC):
    """New Adapter for DB-GPT LLM models"""

    model_name: Optional[str] = None
    model_path: Optional[str] = None
    conv_factory: Optional[ConversationAdapterFactory] = None
    # TODO: more flexible quantization config
    support_4bit: bool = False
    support_8bit: bool = False
    support_system_message: bool = True
    _supported_models: List[ModelMetadata] = []

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} model_name={self.model_name} model_path="
            f"{self.model_path}>"
        )

    def __str__(self):
        return self.__repr__()

    def new_adapter(self, **kwargs) -> "LLMModelAdapter":
        """Create a new adapter instance

        Args:
            **kwargs: The parameters of the new adapter instance

        Returns:
            LLMModelAdapter: The new adapter instance
        """
        new_obj = self.__class__()
        new_obj.model_name = self.model_name
        new_obj.model_path = self.model_path
        new_obj.conv_factory = self.conv_factory
        new_obj.support_4bit = self.support_4bit
        new_obj.support_8bit = self.support_8bit
        new_obj.support_system_message = self.support_system_message
        new_obj._supported_models = self._supported_models
        return new_obj

    def use_fast_tokenizer(self) -> bool:
        """Whether use a [fast Rust-based tokenizer]
        (https://huggingface.co/docs/tokenizers/index) if it is supported
        for a given model.
        """
        return False

    def model_type(self) -> str:
        return ModelType.HF

    def model_param_class(
        self, model_type: str = None
    ) -> Type[LLMDeployModelParameters]:
        """Get the startup parameters instance of the model

        Args:
            model_type (str, optional): The type of model. Defaults to None.

        Returns:
            Type[BaseModelParameters]: The startup parameters instance of the model
        """

    def match(
        self,
        provider: str,
        model_name: Optional[str] = None,
        model_path: Optional[str] = None,
    ) -> bool:
        """Whether the model adapter can load the given model

        Args:
            provider (str): The provider of the model, old version is model_type.
            model_name (Optional[str], optional): The name of model. Defaults to None.
            model_path (Optional[str], optional): The path of model. Defaults to None.
        """
        return False

    def supported_models(self) -> List[ModelMetadata]:
        """Return the supported models."""
        return self._supported_models

    def support_quantization_4bit(self) -> bool:
        """Whether the model adapter can load 4bit model

        If it is True, we will load the 4bit model with :meth:`~LLMModelAdapter.load`

        Returns:
            bool: Whether the model adapter can load 4bit model, default is False
        """
        return self.support_4bit

    def support_quantization_8bit(self) -> bool:
        """Whether the model adapter can load 8bit model

        If it is True, we will load the 8bit model with :meth:`~LLMModelAdapter.load`

        Returns:
            bool: Whether the model adapter can load 8bit model, default is False
        """
        return self.support_8bit

    def load(self, model_path: str, from_pretrained_kwargs: dict):
        """Load model and tokenizer"""
        raise NotImplementedError

    def model_patch(
        self, deploy_model_params: LLMDeployModelParameters
    ) -> Optional[Callable[[Any], Any]]:
        """Patch function for model"""
        return None

    def parse_max_length(self, model, tokenizer) -> Optional[int]:
        """Parse the max_length of the model.

        Returns:
            Optional[int]: The max_length of the model
        """
        if not (tokenizer or model):
            return None
        try:
            model_max_length = None
            if tokenizer and hasattr(tokenizer, "model_max_length"):
                model_max_length = tokenizer.model_max_length
            if model_max_length and model_max_length < 100000000:
                # Can't be too large
                return model_max_length
            if model and hasattr(model, "config"):
                model_config = model.config
                if hasattr(model_config, "max_sequence_length"):
                    return model_config.max_sequence_length
                if hasattr(model_config, "max_position_embeddings"):
                    return model_config.max_position_embeddings
            return None
        except Exception:
            return None

    def load_from_params(self, params):
        """Load the model and tokenizer according to the given parameters"""
        raise NotImplementedError

    def is_reasoning_model(
        self,
        deploy_model_params: LLMDeployModelParameters,
        lower_model_name_or_path: Optional[str] = None,
    ) -> bool:
        """Whether the model is a reasoning model"""
        if (
            deploy_model_params.reasoning_model is not None
            and deploy_model_params.reasoning_model
        ):
            return True
        return (
            (
                lower_model_name_or_path
                and "deepseek" in lower_model_name_or_path
                and (
                    "r1" in lower_model_name_or_path
                    or "reasoning" in lower_model_name_or_path
                    or "reasoner" in lower_model_name_or_path
                )
            )
            or (lower_model_name_or_path and "qwq" in lower_model_name_or_path)
            or (lower_model_name_or_path and "qwen3" in lower_model_name_or_path)
        )

    def support_async(self) -> bool:
        """Whether the loaded model supports asynchronous calls"""
        return False

    def support_generate_function(self) -> bool:
        """Whether support generate function, if it is False, we will use
        generate_stream function.

        Sometimes, we need to use generate function to get the result of the model.
        """
        return False

    def get_generate_stream_function(
        self, model, deploy_model_params: LLMDeployModelParameters
    ):
        """Get the generate stream function of the model"""
        raise NotImplementedError

    def get_async_generate_stream_function(
        self, model, deploy_model_params: LLMDeployModelParameters
    ):
        """Get the asynchronous generate stream function of the model"""
        raise NotImplementedError

    def get_generate_function(
        self, model, deploy_model_params: LLMDeployModelParameters
    ):
        """Get the generate function of the model"""
        raise NotImplementedError

    def get_async_generate_function(self, model, model_path: str):
        """Get the asynchronous generate function of the model"""
        raise NotImplementedError

    def get_media_processor(
        self,
        deploy_model_params: LLMDeployModelParameters,  # noqa: F821
    ) -> Optional[MediaProcessor]:
        """Get the media processor of the model.

        Some models may need to process the media content before sending to the model,
        such as image or audio. This method will be called before sending the media.

        If your model provider is proxy model, you may be upload the local file to the
        public cloud, and then get the url of the file. So you can implement the
        :meth:`~LLMModelAdapter.get_media_processor` method to upload the files.

        Or in cluster mode and the model deploy locally, we must provide the media
        processor to download the files to the local disk.

        Args:
            deploy_model_params (LLMDeployModelParameters): The model parameters.

        Returns:
            Optional[MediaProcessor]: The media processor.
        """
        return None

    def get_default_conv_template(
        self, model_name: str, model_path: str
    ) -> Optional[ConversationAdapter]:
        """Get the default conversation template

        Args:
            model_name (str): The name of the model.
            model_path (str): The path of the model.

        Returns:
            Optional[ConversationAdapter]: The conversation template.
        """
        raise NotImplementedError

    def get_default_message_separator(self) -> str:
        """Get the default message separator"""
        try:
            conv_template = self.get_default_conv_template(
                self.model_name, self.model_path
            )
            return conv_template.sep
        except Exception:
            return "\n"

    def get_prompt_roles(self) -> List[str]:
        """Get the roles of the prompt

        Returns:
            List[str]: The roles of the prompt
        """
        roles = [ModelMessageRoleType.HUMAN, ModelMessageRoleType.AI]
        if self.support_system_message:
            roles.append(ModelMessageRoleType.SYSTEM)
        return roles

    def transform_model_messages(
        self,
        messages: List[ModelMessage],
        convert_to_compatible_format: bool = False,
        support_media_content: bool = True,
        type_mapping: Optional[Dict[str, str]] = None,
    ) -> List[Dict[str, str]]:
        """Transform the model messages

        Default is the OpenAI format, example:
            .. code-block:: python

                return_messages = [
                    {"role": "system", "content": "You are a helpful assistant"},
                    {"role": "user", "content": "Hello"},
                    {"role": "assistant", "content": "Hi"},
                ]

        But some model may need to transform the messages to other format(e.g. There is
         no system message), such as:
            .. code-block:: python

                return_messages = [
                    {"role": "user", "content": "Hello"},
                    {"role": "assistant", "content": "Hi"},
                ]
        Args:
            messages (List[ModelMessage]): The model messages
            convert_to_compatible_format (bool, optional): Whether to convert to
                compatible format. Defaults to False.
            support_media_content (bool, optional): Whether to support media content

        Returns:
            List[Dict[str, str]]: The transformed model messages
        """
        logger.info(f"support_system_message: {self.support_system_message}")
        if not self.support_system_message and convert_to_compatible_format:
            # We will not do any transform in the future
            return self._transform_to_no_system_messages(
                messages,
                support_media_content=support_media_content,
                type_mapping=type_mapping,
            )
        else:
            return ModelMessage.to_common_messages(
                messages,
                convert_to_compatible_format=convert_to_compatible_format,
                support_media_content=support_media_content,
                type_mapping=type_mapping,
            )

    def _transform_to_no_system_messages(
        self,
        messages: List[ModelMessage],
        support_media_content: bool = True,
        type_mapping: Optional[Dict[str, str]] = None,
    ) -> List[Dict[str, str]]:
        """Transform the model messages to no system messages

        Some opensource chat model no system messages, so wo should transform the
        messages to no system messages.

        Merge the system messages to the last user message, example:
            .. code-block:: python

                return_messages = [
                    {"role": "system", "content": "You are a helpful assistant"},
                    {"role": "user", "content": "Hello"},
                    {"role": "assistant", "content": "Hi"},
                ]
            =>
                return_messages = [
                    {"role": "user", "content": "You are a helpful assistant\nHello"},
                    {"role": "assistant", "content": "Hi"}
                ]

        Args:
            messages (List[ModelMessage]): The model messages

        Returns:
            List[Dict[str, str]]: The transformed model messages
        """
        openai_messages = ModelMessage.to_common_messages(
            messages,
            support_media_content=support_media_content,
            type_mapping=type_mapping,
        )
        system_messages = []
        return_messages = []
        for message in openai_messages:
            if message["role"] == "system":
                system_messages.append(message["content"])
            else:
                return_messages.append(message)
        if len(system_messages) > 1:
            # Too much system messages should be a warning
            logger.warning("Your system messages have more than one message")
        if system_messages:
            sep = self.get_default_message_separator()
            str_system_messages = ",".join(system_messages)
            # Update last user message
            return_messages[-1]["content"] = (
                str_system_messages + sep + return_messages[-1]["content"]
            )
        return return_messages

    def get_str_prompt(
        self,
        params: Dict,
        messages: List[ModelMessage],
        tokenizer: Any,
        prompt_template: str = None,
        convert_to_compatible_format: bool = False,
    ) -> Optional[str]:
        """Get the string prompt from the given parameters and messages

        If the value of return is not None, we will skip
        :meth:`~LLMModelAdapter.get_prompt_with_template` and use the value of return.

        Args:
            params (Dict): The parameters
            messages (List[ModelMessage]): The model messages
            tokenizer (Any): The tokenizer of model, in huggingface chat model, we can
                create the prompt by tokenizer
            prompt_template (str, optional): The prompt template. Defaults to None.
            convert_to_compatible_format (bool, optional): Whether to convert to
                compatible format. Defaults to False.

        Returns:
            Optional[str]: The string prompt
        """
        return None

    def load_media(self, params: Dict, messages: List[ModelMessage]):
        """Load the media content from the messages

        Args:
            params (Dict): The parameters
            messages (List[ModelMessage]): The model messages
        """
        pass

    def get_prompt_with_template(
        self,
        params: Dict,
        messages: List[ModelMessage],
        model_name: str,
        model_path: str,
        model_context: Dict,
        prompt_template: str = None,
    ):
        convert_to_compatible_format = params.get("convert_to_compatible_format")
        conv: ConversationAdapter = self.get_default_conv_template(
            model_name, model_path
        )

        if prompt_template:
            logger.info(f"Use prompt template {prompt_template} from config")
            conv = get_conv_template(prompt_template)
        if not conv or not messages:
            # Nothing to do
            logger.info(
                f"No conv from model_path {model_path} or no messages in params, {self}"
            )
            return None, None, None

        conv = conv.copy()
        if convert_to_compatible_format:
            # In old version, we will convert the messages to compatible format
            conv = self._set_conv_converted_messages(conv, messages)
        else:
            # In new version, we will use the messages directly
            conv = self._set_conv_messages(conv, messages)

        # Add a blank message for the assistant.
        conv.append_message(conv.roles[1], None)
        new_prompt = conv.get_prompt()
        return new_prompt, conv.stop_str, conv.stop_token_ids

    def _set_conv_messages(
        self, conv: ConversationAdapter, messages: List[ModelMessage]
    ) -> ConversationAdapter:
        """Set the messages to the conversation template

        Args:
            conv (ConversationAdapter): The conversation template
            messages (List[ModelMessage]): The model messages

        Returns:
            ConversationAdapter: The conversation template with messages
        """
        system_messages = []
        for message in messages:
            if isinstance(message, ModelMessage):
                role = message.role
                content = message.content
            elif isinstance(message, dict):
                role = message["role"]
                content = message["content"]
            else:
                raise ValueError(f"Invalid message type: {message}")

            if role == ModelMessageRoleType.SYSTEM:
                system_messages.append(content)
            elif role == ModelMessageRoleType.HUMAN:
                conv.append_message(conv.roles[0], content)
            elif role == ModelMessageRoleType.AI:
                conv.append_message(conv.roles[1], content)
            else:
                raise ValueError(f"Unknown role: {role}")
        if len(system_messages) > 1:
            raise ValueError(
                f"Your system messages have more than one message: {system_messages}"
            )
        if system_messages:
            conv.set_system_message(system_messages[0])
        return conv

    def _set_conv_converted_messages(
        self, conv: ConversationAdapter, messages: List[ModelMessage]
    ) -> ConversationAdapter:
        """Set the messages to the conversation template

        In the old version, we will convert the messages to compatible format.
        This method will be deprecated in the future.

        Args:
            conv (ConversationAdapter): The conversation template
            messages (List[ModelMessage]): The model messages

        Returns:
            ConversationAdapter: The conversation template with messages
        """
        system_messages = []
        user_messages = []
        ai_messages = []

        for message in messages:
            if isinstance(message, ModelMessage):
                role = message.role
                content = message.content
            elif isinstance(message, dict):
                role = message["role"]
                content = message["content"]
            else:
                raise ValueError(f"Invalid message type: {message}")

            if role == ModelMessageRoleType.SYSTEM:
                # Support for multiple system messages
                system_messages.append(content)
            elif role == ModelMessageRoleType.HUMAN:
                user_messages.append(content)
            elif role == ModelMessageRoleType.AI:
                ai_messages.append(content)
            else:
                raise ValueError(f"Unknown role: {role}")

        can_use_systems: [] = []
        if system_messages:
            if len(system_messages) > 1:
                # Compatible with dbgpt complex scenarios, the last system will protect
                # more complete information
                # entered by the current user
                user_messages[-1] = system_messages[-1]
                can_use_systems = system_messages[:-1]
            else:
                can_use_systems = system_messages

        for i in range(len(user_messages)):
            conv.append_message(conv.roles[0], user_messages[i])
            if i < len(ai_messages):
                conv.append_message(conv.roles[1], ai_messages[i])

        # TODO join all system messages may not be a good idea
        conv.set_system_message("".join(can_use_systems))
        return conv

    def apply_conv_template(self) -> bool:
        return self.model_type() != ModelType.PROXY

    def model_adaptation(
        self,
        params: Dict,
        model_name: str,
        model_path: str,
        tokenizer: Any,
        prompt_template: Optional[str] = None,
    ) -> Tuple[Dict, Dict]:
        """Params adaptation"""
        messages = params.get("messages")
        convert_to_compatible_format = params.get("convert_to_compatible_format")
        message_version = params.get("version", "v2").lower()
        logger.info(f"Message version is {message_version}")
        if convert_to_compatible_format is None:
            # Support convert messages to compatible format when message version is v1
            convert_to_compatible_format = message_version == "v1"
        # Save to params
        params["convert_to_compatible_format"] = convert_to_compatible_format

        # Some model context to dbgpt server
        model_context = {
            "prompt_echo_len_char": -1,
            "has_format_prompt": False,
            "echo": params.get("echo", True),
        }
        if messages:
            # Dict message to ModelMessage
            messages = [
                m if isinstance(m, ModelMessage) else ModelMessage(**m)
                for m in messages
            ]
            params["messages"] = messages
        params["string_prompt"] = ModelMessage.messages_to_string(messages)

        # Load media content to params
        self.load_media(params, messages)

        if not self.apply_conv_template():
            # No need to apply conversation template, now for proxy LLM
            return params, model_context

        new_prompt = self.get_str_prompt(
            params, messages, tokenizer, prompt_template, convert_to_compatible_format
        )
        conv_stop_str, conv_stop_token_ids = None, None
        if not new_prompt:
            (
                new_prompt,
                conv_stop_str,
                conv_stop_token_ids,
            ) = self.get_prompt_with_template(
                params, messages, model_name, model_path, model_context, prompt_template
            )
            if not new_prompt:
                return params, model_context

        # Overwrite the original prompt
        # TODO remote bos token and eos token from tokenizer_config.json of model
        prompt_echo_len_char = len(new_prompt.replace("</s>", "").replace("<s>", ""))
        model_context["prompt_echo_len_char"] = prompt_echo_len_char
        model_context["has_format_prompt"] = True
        params["prompt"] = new_prompt

        custom_stop = params.get("stop")
        custom_stop_token_ids = params.get("stop_token_ids")

        # Prefer the value passed in from the input parameter
        params["stop"] = custom_stop or conv_stop_str
        params["stop_token_ids"] = custom_stop_token_ids or conv_stop_token_ids

        return params, model_context


class AdapterEntry:
    """The entry of model adapter"""

    def __init__(
        self,
        model_adapter: Union[LLMModelAdapter, EmbeddingModelAdapter],
        match_funcs: List[Callable[[str, str, str], bool]] = None,
    ):
        self.model_adapter = model_adapter
        self.match_funcs = match_funcs or []

    def __repr__(self) -> str:
        return f"<AdapterEntry model_adapter={self.model_adapter}>"


model_adapters: List[AdapterEntry] = []
embedding_adapters: List[AdapterEntry] = []


def register_model_adapter(
    model_adapter_cls: Type[LLMModelAdapter],
    match_funcs: List[Callable[[str, str, str], bool]] = None,
    supported_models: List[ModelMetadata] = None,
) -> None:
    """Register a model adapter.

    Args:
        model_adapter_cls (Type[LLMModelAdapter]): The model adapter class.
        match_funcs (List[Callable[[str, str, str], bool]], optional): The match
            functions. Defaults to None.
        supported_models (List[ModelMetadata], optional): The supported models.
    """
    adapter = model_adapter_cls()
    if supported_models:
        adapter._supported_models = supported_models
    model_adapters.append(AdapterEntry(adapter, match_funcs))


def get_model_adapter(
    provider: str,
    model_name: Optional[str] = None,
    model_path: Optional[str] = None,
    conv_factory: Optional[ConversationAdapterFactory] = None,
) -> Optional[LLMModelAdapter]:
    """Get a model adapter.

    Args:
        provider (str): Provider of the model, old version is model_type.
        model_name (Optional[str], optional): The name of the model. Defaults to None.
        model_path (Optional[str], optional): The path of the model. Defaults to None.
        conv_factory (Optional[ConversationAdapterFactory], optional): The conversation
         factory. Defaults to None.
    Returns:
        Optional[LLMModelAdapter]: The model adapter.
    """
    adapter = None
    # First find adapter by model type
    adapters_by_provider = []
    for adapter_entry in model_adapters[::-1]:
        if adapter_entry.model_adapter.match(provider, None, None):
            adapters_by_provider.append(adapter_entry)
    if adapters_by_provider and len(adapters_by_provider) == 1:
        adapter = adapters_by_provider[0].model_adapter

    # Then find adapter by model name
    for adapter_entry in model_adapters[::-1]:
        if adapter_entry.model_adapter.match(provider, model_name, None):
            adapter = adapter_entry.model_adapter
            break
    for adapter_entry in model_adapters[::-1]:
        if adapter_entry.model_adapter.match(provider, None, model_path):
            adapter = adapter_entry.model_adapter
            break
    if adapter:
        new_adapter = adapter.new_adapter()
        new_adapter.model_name = model_name
        new_adapter.model_path = model_path
        if conv_factory:
            new_adapter.conv_factory = conv_factory
        return new_adapter
    return None


def get_supported_models(worker_type: str) -> List[SupportedModel]:
    """Get the supported models."""
    models = []
    from dbgpt.util.parameter_utils import _get_parameter_descriptions

    adapters = (
        model_adapters if worker_type == WorkerType.LLM.value else embedding_adapters
    )

    for adapter_entry in adapters[::-1]:
        model_adapter = adapter_entry.model_adapter
        try:
            params_cls = model_adapter.model_param_class()
            if params_cls.worker_type() != worker_type:
                continue
            provider = params_cls.provider or model_adapter.model_type()
            is_proxy = provider == ModelType.PROXY or provider.startswith(
                ModelType.PROXY
            )
            for m in model_adapter.supported_models():
                real_models = m.model if isinstance(m.model, list) else [m.model]
                for model_name in real_models:
                    doc_title = m.label or model_name
                    description = m.description or ""
                    if m.link:
                        description += f"\n[More Details]({m.link})"
                    description = "## " + doc_title + "\n\n" + description
                    sm = SupportedModel(
                        model=model_name,
                        worker_type=worker_type,
                        provider=provider,
                        proxy=is_proxy,
                        enabled=True,
                        params=_get_parameter_descriptions(params_cls),
                        description=description,
                    )
                    models.append(sm)
        except Exception as e:
            logger.warning(f"{model_adapter} get supported models failed: {e}")
    return models


def register_embedding_adapter(
    model_adapter_cls: Union[
        Type[EmbeddingModelAdapter], Type[Embeddings], Type[RerankEmbeddings]
    ],
    supported_models: List[EmbeddingModelMetadata],
    match_funcs: List[Callable[[str, str, str], bool]] = None,
) -> None:
    """Register an embedding adapter.

    Args:
        model_adapter_cls (Type[EmbeddingModelAdapter]): The embedding adapter class.
        supported_models (List[EmbeddingModelMetadata]): The supported models.
        match_funcs (List[Callable[[str, str, str], bool]], optional): The match
            functions. Defaults to None.
    """
    if issubclass(model_adapter_cls, Embeddings) or issubclass(
        model_adapter_cls, RerankEmbeddings
    ):
        params_adapter_cls = model_adapter_cls.param_class()
        provider = params_adapter_cls.provider
        match_fun = model_adapter_cls._match
        is_rerank_params = issubclass(model_adapter_cls, RerankEmbeddings)

        class _DyEmbeddingAdapter(EmbeddingModelAdapter):
            def model_param_class(self, model_type: str = None):
                return params_adapter_cls

            def supported_models(self):
                return supported_models

            def do_match(
                self, is_rerank: bool, lower_model_name_or_path: Optional[str] = None
            ):
                if is_rerank_params != is_rerank:
                    return False
                return match_fun(provider, lower_model_name_or_path)

            def load_from_params(self, params):
                return model_adapter_cls.from_parameters(params)

        embedding_adapters.append(AdapterEntry(_DyEmbeddingAdapter(), match_funcs))
    else:
        embedding_adapters.append(AdapterEntry(model_adapter_cls(), match_funcs))


def get_embedding_adapter(
    provider: str,
    is_rerank: bool,
    model_name: Optional[str] = None,
    model_path: Optional[str] = None,
) -> Optional[EmbeddingModelAdapter]:
    """Get an embedding adapter.

    Args:
        provider (str): Provider of the model, old version is model_type.
        is_rerank (bool): Whether the model is a rerank model.
        model_name (Optional[str], optional): The name of the model. Defaults to None.
        model_path (Optional[str], optional): The path of the model. Defaults to None.
    Returns:
        Optional[EmbeddingModelAdapter]: The embedding adapter.
    """
    adapter: Optional[EmbeddingModelAdapter] = None
    # First find adapter by model type
    adapters_by_provider = []
    for adapter_entry in embedding_adapters[::-1]:
        if adapter_entry.model_adapter.match(provider, is_rerank, None, None):
            adapters_by_provider.append(adapter_entry)
    if adapters_by_provider and len(adapters_by_provider) == 1:
        adapter = adapters_by_provider[0].model_adapter

    # Then find adapter by model name
    for adapter_entry in embedding_adapters[::-1]:
        if adapter_entry.model_adapter.match(provider, is_rerank, model_name, None):
            adapter = adapter_entry.model_adapter
            break
    for adapter_entry in embedding_adapters[::-1]:
        if adapter_entry.model_adapter.match(provider, is_rerank, None, model_path):
            adapter = adapter_entry.model_adapter
            break
    if adapter:
        new_adapter = adapter.new_adapter()
        new_adapter.model_name = model_name
        new_adapter.model_path = model_path
        return new_adapter
    return None



from typing import AsyncIterator

import pytest

from dbgpt.core.awel import (
    DAG,
    InputSource,
    MapOperator,
    StreamifyAbsOperator,
    TransformStreamAbsOperator,
)
from dbgpt.core.awel.trigger.iterator_trigger import IteratorTrigger


class NumberProducerOperator(StreamifyAbsOperator[int, int]):
    """Create a stream of numbers from 0 to `n-1`"""

    async def streamify(self, n: int) -> AsyncIterator[int]:
        for i in range(n):
            yield i


class MyStreamingOperator(TransformStreamAbsOperator[int, int]):
    async def transform_stream(self, data: AsyncIterator[int]) -> AsyncIterator[int]:
        async for i in data:
            yield i * i


async def _check_stream_results(stream_results, expected_len):
    assert len(stream_results) == expected_len
    for _, result in stream_results:
        i = 0
        async for num in result:
            assert num == i * i
            i += 1


@pytest.mark.asyncio
async def test_single_data():
    with DAG("test_single_data"):
        trigger_task = IteratorTrigger(data=2)
        task = MapOperator(lambda x: x * x)
        trigger_task >> task
    results = await trigger_task.trigger()
    assert len(results) == 1
    assert results[0][1] == 4

    with DAG("test_single_data_stream"):
        trigger_task = IteratorTrigger(data=2, streaming_call=True)
        number_task = NumberProducerOperator()
        task = MyStreamingOperator()
        trigger_task >> number_task >> task
    stream_results = await trigger_task.trigger()
    await _check_stream_results(stream_results, 1)


@pytest.mark.asyncio
async def test_list_data():
    with DAG("test_list_data"):
        trigger_task = IteratorTrigger(data=[0, 1, 2, 3])
        task = MapOperator(lambda x: x * x)
        trigger_task >> task
    results = await trigger_task.trigger()
    assert len(results) == 4
    assert results == [(0, 0), (1, 1), (2, 4), (3, 9)]

    with DAG("test_list_data_stream"):
        trigger_task = IteratorTrigger(data=[0, 1, 2, 3], streaming_call=True)
        number_task = NumberProducerOperator()
        task = MyStreamingOperator()
        trigger_task >> number_task >> task
    stream_results = await trigger_task.trigger()
    await _check_stream_results(stream_results, 4)


@pytest.mark.asyncio
async def test_async_iterator_data():
    async def async_iter():
        for i in range(4):
            yield i

    with DAG("test_async_iterator_data"):
        trigger_task = IteratorTrigger(data=async_iter())
        task = MapOperator(lambda x: x * x)
        trigger_task >> task
    results = await trigger_task.trigger()
    assert len(results) == 4
    assert results == [(0, 0), (1, 1), (2, 4), (3, 9)]

    with DAG("test_async_iterator_data_stream"):
        trigger_task = IteratorTrigger(data=async_iter(), streaming_call=True)
        number_task = NumberProducerOperator()
        task = MyStreamingOperator()
        trigger_task >> number_task >> task
    stream_results = await trigger_task.trigger()
    await _check_stream_results(stream_results, 4)


@pytest.mark.asyncio
async def test_input_source_data():
    with DAG("test_input_source_data"):
        trigger_task = IteratorTrigger(data=InputSource.from_iterable([0, 1, 2, 3]))
        task = MapOperator(lambda x: x * x)
        trigger_task >> task
    results = await trigger_task.trigger()
    assert len(results) == 4
    assert results == [(0, 0), (1, 1), (2, 4), (3, 9)]

    with DAG("test_input_source_data_stream"):
        trigger_task = IteratorTrigger(
            data=InputSource.from_iterable([0, 1, 2, 3]),
            streaming_call=True,
        )
        number_task = NumberProducerOperator()
        task = MyStreamingOperator()
        trigger_task >> number_task >> task
    stream_results = await trigger_task.trigger()
    await _check_stream_results(stream_results, 4)


@pytest.mark.asyncio
async def test_parallel_safe():
    with DAG("test_parallel_safe"):
        trigger_task = IteratorTrigger(data=InputSource.from_iterable([0, 1, 2, 3]))
        task = MapOperator(lambda x: x * x)
        trigger_task >> task
    results = await trigger_task.trigger(parallel_num=3)
    assert len(results) == 4
    assert results == [(0, 0), (1, 1), (2, 4), (3, 9)]

    with DAG("test_input_source_data_stream"):
        trigger_task = IteratorTrigger(
            data=InputSource.from_iterable([0, 1, 2, 3]),
            streaming_call=True,
        )
        number_task = NumberProducerOperator()
        task = MyStreamingOperator()
        trigger_task >> number_task >> task
    stream_results = await trigger_task.trigger(parallel_num=3)
    await _check_stream_results(stream_results, 4)



from typing import Dict

import mlx.nn as nn
from mlx_lm import stream_generate
from mlx_lm.sample_utils import make_sampler
from mlx_lm.tokenizer_utils import TokenizerWrapper

from dbgpt.core import ModelOutput

from ...utils.llm_metrics import LLMPerformanceMonitor
from ...utils.parse_utils import (
    _DEFAULT_THINK_END_TOKEN,
    _DEFAULT_THINK_START_TOKEN,
    parse_chat_message,
)


def generate_stream(
    model: nn.Module,
    tokenizer: TokenizerWrapper,
    params: Dict,
    device: str,
    context_len: int,
):
    prompt = params["prompt"]
    temperature = float(params.get("temperature", 0))
    top_p = float(params.get("top_p", 1.0))
    top_k = params.get("top_k", 0)
    max_new_tokens = int(params.get("max_new_tokens", 2048))
    # echo = bool(params.get("echo", True))
    think_start_token = params.get("think_start_token", _DEFAULT_THINK_START_TOKEN)
    think_end_token = params.get("think_end_token", _DEFAULT_THINK_END_TOKEN)
    is_reasoning_model = params.get("is_reasoning_model", False)

    reasoning_patterns = [
        {"start": think_start_token, "end": think_end_token},
    ]
    sampler = make_sampler(
        temp=temperature,
        top_p=top_p,
        # min_p=min_p,
        # min_tokens_to_keep=min_tokens_to_keep,
        top_k=top_k,
        xtc_special_tokens=tokenizer.encode("\n") + list(tokenizer.eos_token_ids),
    )

    # Initialize the performance monitor with estimated token count
    estimated_input_tokens = len(tokenizer.encode(prompt))
    perf_monitor = LLMPerformanceMonitor(input_token_count=estimated_input_tokens)

    # Start measuring prefill phase
    perf_monitor.start_prefill()

    results_generator = stream_generate(
        model, tokenizer, prompt=prompt, max_tokens=max_new_tokens, sampler=sampler
    )
    text = ""
    is_first = True
    for res in results_generator:
        new_text = res.text
        text += new_text
        # The prompt processing tokens-per-second.
        # prompt_tps = res.prompt_tps
        # The number of tokens in the prompt.
        prompt_tokens = res.prompt_tokens
        # The number of generated tokens.
        generation_tokens = res.generation_tokens
        # The tokens-per-second for generation.
        # generation_tps = res.generation_tps
        # The peak memory used so far in GB.
        # peak_memory = res.peak_memory
        # "length", "stop" or `None`
        finish_reason = res.finish_reason
        if (
            prompt.rstrip().endswith(think_start_token)
            and is_reasoning_model
            and is_first
        ):
            text = think_start_token + "\n" + text
            is_first = False

        msg = parse_chat_message(
            text,
            extract_reasoning=is_reasoning_model,
            reasoning_patterns=reasoning_patterns,
        )

        # If this is the first iteration, update the input token count
        if perf_monitor.metrics.input_token_count != prompt_tokens:
            perf_monitor.metrics.input_token_count = prompt_tokens
        # Update performance metrics based on current token count
        perf_metrics = perf_monitor.on_tokens_received(generation_tokens)
        usage = {
            "prompt_tokens": prompt_tokens,
            "completion_tokens": generation_tokens,
            "total_tokens": prompt_tokens + generation_tokens,
        }
        # Check if generation is complete
        is_complete = finish_reason is not None
        if is_complete:
            perf_monitor.end_generation()
        usage.update(perf_metrics)

        yield ModelOutput.build(
            msg.content,
            msg.reasoning_content,
            error_code=0,
            usage=usage,
            is_reasoning_model=is_reasoning_model,
        )



import pytest

from dbgpt.util.sql_utils import remove_sql_comments


class TestRemoveSqlComments:
    """Test cases for the remove_sql_comments method."""

    def test_remove_single_line_comments_basic(self):
        """Test removing basic single-line comments."""
        sql = "SELECT * FROM users -- This is a comment"
        expected = "SELECT * FROM users "
        result = remove_sql_comments(sql)
        assert result == expected

    def test_remove_single_line_comments_with_newline(self):
        """Test removing single-line comments followed by newline."""
        sql = "SELECT * FROM users -- This is a comment\nWHERE id = 1"
        expected = "SELECT * FROM users WHERE id = 1"
        result = remove_sql_comments(sql)
        assert result == expected

    def test_remove_single_line_comments_at_end(self):
        """Test removing single-line comments at the end of string."""
        sql = "SELECT * FROM users\n-- Final comment"
        expected = "SELECT * FROM users\n"
        result = remove_sql_comments(sql)
        assert result == expected

    def test_remove_multiple_single_line_comments(self):
        """Test removing multiple single-line comments."""
        sql = """SELECT * FROM users -- Comment 1
                 WHERE id = 1 -- Comment 2
                 ORDER BY name -- Comment 3"""
        expected = """SELECT * FROM users                  WHERE id = 1                  ORDER BY name """  # noqa: E501
        result = remove_sql_comments(sql)
        assert result == expected

    def test_remove_multiline_comments_basic(self):
        """Test removing basic multi-line comments."""
        sql = "SELECT * FROM users /* This is a comment */ WHERE id = 1"
        expected = "SELECT * FROM users  WHERE id = 1"
        result = remove_sql_comments(sql)
        assert result == expected

    def test_remove_multiline_comments_with_newlines(self):
        """Test removing multi-line comments spanning multiple lines."""
        sql = """SELECT * FROM users 
                 /* This is a 
                    multi-line comment */ 
                 WHERE id = 1"""
        expected = """SELECT * FROM users 
                  
                 WHERE id = 1"""
        result = remove_sql_comments(sql)
        assert result == expected

    def test_remove_multiple_multiline_comments(self):
        """Test removing multiple multi-line comments."""
        sql = "SELECT * /* comment 1 */ FROM users /* comment 2 */ WHERE id = 1"
        expected = "SELECT *  FROM users  WHERE id = 1"
        result = remove_sql_comments(sql)
        assert result == expected

    def test_remove_mixed_comments(self):
        """Test removing both single-line and multi-line comments."""
        sql = """SELECT * FROM users /* multi-line comment */
                 WHERE id = 1 -- single line comment
                 /* another multi-line 
                    comment */
                 ORDER BY name -- final comment"""
        expected = """SELECT * FROM users 
                 WHERE id = 1                  
                 ORDER BY name """
        result = remove_sql_comments(sql)
        assert result == expected

    def test_no_comments_to_remove(self):
        """Test SQL with no comments."""
        sql = "SELECT * FROM users WHERE id = 1 ORDER BY name"
        expected = "SELECT * FROM users WHERE id = 1 ORDER BY name"
        result = remove_sql_comments(sql)
        assert result == expected

    def test_empty_string(self):
        """Test with empty string."""
        sql = ""
        expected = ""
        result = remove_sql_comments(sql)
        assert result == expected

    def test_only_comments(self):
        """Test string with only comments."""
        sql = "-- This is only a comment"
        expected = ""
        result = remove_sql_comments(sql)
        assert result == expected

    def test_only_multiline_comments(self):
        """Test string with only multi-line comments."""
        sql = "/* This is only a multi-line comment */"
        expected = ""
        result = remove_sql_comments(sql)
        assert result == expected

    def test_comments_with_special_characters(self):
        """Test comments containing special characters."""
        sql = "SELECT * FROM users -- Comment with special chars: @#$%^&*()"
        expected = "SELECT * FROM users "
        result = remove_sql_comments(sql)
        assert result == expected

    def test_comments_with_sql_keywords(self):
        """Test comments containing SQL keywords."""
        sql = "SELECT * FROM users -- SELECT * FROM another_table"
        expected = "SELECT * FROM users "
        result = remove_sql_comments(sql)
        assert result == expected

    def test_whitespace_handling(self):
        """Test whitespace handling around comments."""
        sql = "SELECT *    -- comment with spaces   \n    FROM users"
        expected = "SELECT *        FROM users"
        result = remove_sql_comments(sql)
        assert result == expected

    def test_multiline_comment_at_start(self):
        """Test multi-line comment at the beginning."""
        sql = "/* Initial comment */ SELECT * FROM users"
        expected = " SELECT * FROM users"
        result = remove_sql_comments(sql)
        assert result == expected

    def test_multiline_comment_at_end(self):
        """Test multi-line comment at the end."""
        sql = "SELECT * FROM users /* Final comment */"
        expected = "SELECT * FROM users "
        result = remove_sql_comments(sql)
        assert result == expected

    def test_complex_sql_with_comments(self):
        """Test complex SQL statement with various comment types."""
        sql = """
        /* Query to get user information */
        SELECT 
            u.id, -- User ID
            u.name, /* User full name */
            u.email -- Contact email
        FROM users u
        /* Join with posts table */
        LEFT JOIN posts p ON u.id = p.user_id
        WHERE u.active = 1 -- Only active users
        /* Order by creation date */
        ORDER BY u.created_at DESC
        -- Limit results
        LIMIT 10;
        """

        result = remove_sql_comments(sql)

        # Verify that all comments are removed
        assert "--" not in result
        assert "/*" not in result
        assert "*/" not in result

        # Verify that SQL keywords are preserved
        assert "SELECT" in result
        assert "FROM" in result
        assert "WHERE" in result
        assert "ORDER BY" in result
        assert "LIMIT" in result


# 如果需要测试实际的类实例，可以使用以下参数化测试
@pytest.mark.parametrize(
    "sql_input,expected_output",
    [
        ("SELECT * FROM users", "SELECT * FROM users"),
        ("SELECT * FROM users -- comment", "SELECT * FROM users "),
        ("SELECT * /* comment */ FROM users", "SELECT *  FROM users"),
        ("", ""),
        ("-- only comment", ""),
        ("/* only comment */", ""),
    ],
)
def test_remove_sql_comments_parametrized(sql_input, expected_output):
    """Parametrized test for common cases."""

    # 创建测试实例
    class MockClass:
        def _remove_sql_comments(self, sql: str) -> str:
            import re

            sql = re.sub(r"--.*?(\n|$)", "", sql)
            sql = re.sub(r"/\*.*?\*/", "", sql, flags=re.DOTALL)
            return sql

    instance = MockClass()
    result = instance._remove_sql_comments(sql_input)
    assert result == expected_output



from ..base import Vis


class VisThinking(Vis):
    """VisThinking."""

    @classmethod
    async def build_message(cls, message: str) -> str:
        vis = VisThinking()
        return f"``````{vis.vis_tag()}\n{message}\n``````"

    def sync_display(self, **kwargs) -> str:
        """Display the content using the vis protocol."""
        content = kwargs.get("content")
        return f"``````{self.vis_tag()}\n{content}\n``````"

    @classmethod
    def vis_tag(cls):
        """Vis tag name.

        Returns:
            str: The tag name associated with the visualization.
        """
        return "vis-thinking"



"""Database resource module."""

import dataclasses
import logging
from concurrent.futures import Executor, ThreadPoolExecutor
from typing import Any, Dict, Generic, List, Optional, Tuple, Union

import cachetools

from dbgpt.datasource.rdbms.base import RDBMSConnector
from dbgpt.util.cache_utils import cached
from dbgpt.util.executor_utils import blocking_func_to_async

from .base import P, Resource, ResourceParameters, ResourceType

logger = logging.getLogger(__name__)

_DEFAULT_PROMPT_TEMPLATE = (
    "Database type: {db_type}, related table structure definition: {schemas}"
)
_DEFAULT_PROMPT_TEMPLATE_ZH = "数据库类型：{db_type}，相关表结构定义：{schemas}"


@dataclasses.dataclass
class DBParameters(ResourceParameters):
    """DB parameters class."""

    db_name: str = dataclasses.field(metadata={"help": "DB name"})


class DBResource(Resource[P], Generic[P]):
    """Database resource class."""

    def __init__(
        self,
        name: str,
        db_type: Optional[str] = None,
        db_name: Optional[str] = None,
        dialect: Optional[str] = None,
        executor: Optional[Executor] = None,
        prompt_template: str = _DEFAULT_PROMPT_TEMPLATE,
        **kwargs,
    ):
        """Initialize the DB resource."""
        self._name = name
        self._db_type = db_type
        self._db_name = db_name
        self._dialect = dialect or db_type
        # Executor for running async tasks
        self._executor = executor or ThreadPoolExecutor()
        self._prompt_template = prompt_template

    @classmethod
    def type(cls) -> ResourceType:
        """Return the resource type."""
        return ResourceType.DB

    @property
    def name(self) -> str:
        """Return the resource name."""
        return self._name

    @property
    def db_type(self) -> str:
        """Return the resource name."""
        if not self._db_type:
            raise ValueError("Database type is not set.")
        return self._db_type

    @property
    def dialect(self) -> str:
        """Return the resource name."""
        if not self._dialect:
            raise ValueError("Dialect is not set.")
        return self._dialect

    @cached(cachetools.TTLCache(maxsize=100, ttl=10))
    async def get_prompt(
        self,
        *,
        lang: str = "en",
        prompt_type: str = "default",
        question: Optional[str] = None,
        resource_name: Optional[str] = None,
        **kwargs,
    ) -> Tuple[str, Optional[List[Dict]]]:
        """Get the prompt."""
        if not self._db_name:
            return "No database name provided.", None
        schema_info = await blocking_func_to_async(
            self._executor, self.get_schema_link, db=self._db_name, question=question
        )
        return (
            self._prompt_template.format(db_type=self._db_type, schemas=schema_info),
            None,
        )

    def execute(self, *args, resource_name: Optional[str] = None, **kwargs) -> Any:
        """Execute the resource."""
        copy_kwargs = kwargs.copy()
        if "db" not in copy_kwargs:
            copy_kwargs["db"] = self._db_name
        return self._sync_query(*args, **copy_kwargs)

    async def async_execute(
        self, *args, resource_name: Optional[str] = None, **kwargs
    ) -> Any:
        """Execute the resource asynchronously."""
        copy_kwargs = kwargs.copy()
        if "db" not in copy_kwargs:
            copy_kwargs["db"] = self._db_name
        return await self.query(*args, **copy_kwargs)

    @property
    def is_async(self) -> bool:
        """Return whether the resource is asynchronous."""
        return True

    def get_schema_link(
        self, db: str, question: Optional[str] = None
    ) -> Union[str, List[str]]:
        """Return the schema link of the database."""
        raise NotImplementedError("The run method should be implemented in a subclass.")

    async def query_to_df(self, sql: str, db: Optional[str] = None):
        """Return the query result as a DataFrame."""
        import pandas as pd

        field_names, result = await self.query(sql, db=db)
        return pd.DataFrame(result, columns=field_names)

    async def query(self, sql: str, db: Optional[str] = None):
        """Return the query result."""
        db_name = db or self._db_name
        return await blocking_func_to_async(
            self._executor, self._sync_query, db=db_name, sql=sql
        )

    def _sync_query(self, db: str, sql: str):
        """Return the query result."""
        raise NotImplementedError("The run method should be implemented in a subclass.")


class RDBMSConnectorResource(DBResource[DBParameters]):
    """Connector resource class."""

    def __init__(
        self,
        name: str,
        connector: Optional[RDBMSConnector] = None,
        db_name: Optional[str] = None,
        db_type: Optional[str] = None,
        dialect: Optional[str] = None,
        executor: Optional[Executor] = None,
        **kwargs,
    ):
        """Initialize the connector resource."""
        if not db_type and connector:
            db_type = connector.db_type
        if not dialect and connector:
            dialect = connector.dialect
        if not db_name and connector:
            db_name = connector.get_current_db_name()
        self._connector = connector
        super().__init__(
            name,
            db_type=db_type,
            db_name=db_name,
            dialect=dialect,
            executor=executor,
            **kwargs,
        )

    @property
    def connector(self) -> RDBMSConnector:
        """Return the connector."""
        if not self._connector:
            raise ValueError("Connector is not set.")
        return self._connector

    def get_schema_link(
        self, db: str, question: Optional[str] = None
    ) -> Union[str, List[str]]:
        """Return the schema link of the database."""
        from dbgpt_ext.rag.summary.rdbms_db_summary import _parse_db_summary

        return _parse_db_summary(self.connector)

    def _sync_query(self, db: str, sql: str) -> Tuple[Tuple, List]:
        """Return the query result."""
        result_lst = self.connector.run(sql)
        columns = result_lst[0]
        values = result_lst[1:]
        return columns, values


class SQLiteDBResource(RDBMSConnectorResource):
    """SQLite database resource class."""

    def __init__(
        self, name: str, db_name: str, executor: Optional[Executor] = None, **kwargs
    ):
        """Initialize the SQLite database resource."""
        from dbgpt_ext.datasource.rdbms.conn_sqlite import SQLiteConnector

        conn = SQLiteConnector.from_file_path(db_name)
        super().__init__(name, conn, executor=executor, **kwargs)
